import os
import sys
import Tigger
import numpy
import tempfile

from astLib.astWCS import WCS
from Tigger.Models import SkyModel, ModelClasses

sys.path.append('/scratch/stimela')

utils = __import__('utils')


CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
MSDIR = os.environ["MSDIR"]
OUTPUT = os.environ["OUTPUT"]

cab = utils.readJson(CONFIG)
args = []
msname = None

sofia_file = 'sofia_parameters.par'
wstd = open(sofia_file, 'w')

wstd.write('writeCat.outputDir={:s}\n'.format(OUTPUT))
port2tigger = False
image = None
writecat = False
parameterise = False

for param in cab['parameters']:
    name = param['name']
    value = param['value']

    if value is None:
        continue
    if name == "port2tigger":
        port2tigger = value
        continue
    if name == "steps.doWriteCat":
        writecat = value
    if name == "steps.doParameterise":
        parameterise = value
    if name == "import.inFile":
        image = value

    wstd.write('{0}={1}\n'.format(name, value))

wstd.close()

utils.xrun('sofia_pipeline.py', [sofia_file])

if not port2tigger:
    sys.exit(0)

# convert to data file to Tigger LSM
# First make dummy tigger model
tfile = tempfile.NamedTemporaryFile(suffix='.txt')
tfile.flush()

if image and writecat and parameterise:
    pass
else:
    sys.exit(0)

prefix = os.path.splitext(image)[0]
tname_lsm = prefix + ".lsm.html"
with open(tfile.name, "w") as stdw:
    stdw.write("#format:name ra_d dec_d i emaj_s emin_s pa_d\n")

model = Tigger.load(tfile.name)
tfile.close()


def tigger_src(src, idx):

    name = "SRC%d" % idx
    flux = ModelClasses.Polarization(src["f_int"], 0, 0, 0)
    ra = numpy.deg2rad(src["ra"])
    dec = numpy.deg2rad(src["dec"])
    pos = ModelClasses.Position(ra, dec)
    ex = numpy.deg2rad(src["ell_maj"])
    ey = numpy.deg2rad(src["ell_min"])
    pa = numpy.deg2rad(src["ell_pa"])

    if ex and ey:
        shape = ModelClasses.Gaussian(ex, ey, pa)
    else:
        shape = None
    source = SkyModel.Source(name, pos, flux, shape=shape)
    # Adding source peak flux (error) as extra flux attributes for sources,
    # and to avoid null values for point sources I_peak = src["Total_flux"]
    if shape:
        source.setAttribute("I_peak", float(src["f_peak"]))
    else:
        source.setAttribute("I_peak", float(src["f_int"]))

    return source


with open('{0}_cat.ascii'.format(prefix)) as stdr:
    # Header
    stdr.readline()
    # Column names
    names = stdr.readline().split("#")[1].strip().split()
    # Units
    stdr.readline()
    # Column numbers
    stdr.readline()
    sys.stdout.write(" ".join(names))
    data = numpy.genfromtxt(stdr,
                            names=names + ["col"])

for i, src in enumerate(data):
    model.sources.append(tigger_src(src, i))

wcs = WCS(image)
centre = wcs.getCentreWCSCoords()
model.ra0, model.dec0 = map(numpy.deg2rad, centre)

model.save(tname_lsm)
# Rename using CORPAT
utils.xrun("tigger-convert", [tname_lsm, "--rename -f"])



import os
import sys
import drivecasa
import logging
casa = drivecasa.Casapy(log2term=True, echo_to_stdout=True, timeout=24*3600*10)

sys.path.append("/scratch/stimela")

utils = __import__('utils')

CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
OUTPUT = os.environ["OUTPUT"]
MSDIR = os.environ["MSDIR"]

cab = utils.readJson(CONFIG)

args = {}
for param in cab['parameters']:
    name = param['name']
    value = param['value']

    if value is None:
        continue

    args[name] = value

script = ['{0}(**{1})'.format(cab['binary'], args)]


def log2term(result):
    if result[1]:
        err = '\n'.join(result[1] if result[1] else [''])
        failed = err.lower().find('an error occurred running task') >= 0
        if failed:
            raise RuntimeError('CASA Task failed. See error message above')
        sys.stdout.write('WARNING:: SEVERE messages from CASA run')


result = casa.run_script(script, raise_on_severe=False)
log2term(result)



import os
import sys

sys.path.append("/scratch/stimela")

utils = __import__("utils")

CONFIG = os.environ["CONFIG"]
INDIR = os.environ["INPUT"]
MSDIR = os.environ["MSDIR"]

cab = utils.readJson(CONFIG)

params = cab["parameters"]

_positional = ['antenna-file']
positional = []
for item in _positional:
    param = filter(lambda a: a['name'] == item, params)[0]
    value = param['value']
    params.remove(param)

    if value:
        positional.append(value)

args = []
for param in params:
    key = param['name']
    value = param['value']
    if value in [None, False]:
        continue
    if value is True:
        arg = '{0}{1}'.format(cab['prefix'], key)
    elif hasattr(value, '__iter__'):
        arg = ' '.join(['{0}{1} {2}'.format(cab['prefix'], key, val)
                        for val in value])
    else:
        if isinstance(value, str):
            value = '"{}"'.format(value)
        arg = '{0}{1} {2}'.format(cab['prefix'], key, value)
    args.append(arg)

utils.xrun(cab['binary'], args+['--nolog']+positional)



import sys
import os
import json

sys.path.append("/scratch/stimela")

utils = __import__('utils')

CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
OUTPUT = os.environ["OUTPUT"]
MSDIR = os.environ["MSDIR"]

cab = utils.readJson(CONFIG)

args = {}
tasksuite = None
for param in cab['parameters']:
    name = param['name']
    value = param['value']

    if value is None:
        continue

    args[name] = value

kwargs = "'{}'".format(json.dumps(args))

ARGS = ["flag_phase_drifts",
        "-s antenna_mod",
        kwargs]

utils.xrun(cab['binary'], ARGS)



import os
import sys

sys.path.append('/scratch/stimela')

utils = __import__('utils')


CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
MSDIR = os.environ["MSDIR"]

cab = utils.readJson(CONFIG)
args = []
msname = None

custom_script = "print(\"Nothing has been done\")"

for param in cab['parameters']:
    name = param['name']
    value = param['value']
    if name == "script":
        custom_script = value
        continue
    if value is None:
        continue
    elif value is False:
        continue
    elif value is True:
        value = ''
    args += ['{0}{1} {2}'.format(cab['prefix'], name, value)]

with open("casajob.py.last", "w") as f:
    f.write(custom_script)
utils.xrun(cab['binary'], ["-c", "casajob.py.last"] + args)



import os
import sys
import time
import stimela
from stimela import docker, singularity, udocker, utils, cargo, podman
from stimela.cargo import cab
import logging
import inspect
import re
from stimela.dismissable import dismissable
from future.utils import raise_
from stimela.main import get_cabs

version = stimela.__version__
USER = os.environ["USER"]
UID = os.getuid()
GID = os.getgid()
CAB_PATH = os.path.abspath(os.path.dirname(cab.__file__))

CONT_IO = {
    "docker": {
        "input": "/input",
        "output": "/home/{0:s}/output".format(USER),
        "msfile": "/home/{0:s}/msdir".format(USER),
    },
    "podman": {
        "input": "/input",
        "output": "/home/{0:s}/output".format(USER),
        "msfile": "/home/{0:s}/msdir".format(USER),
    },
    "udocker": {
        "input": "/scratch/input",
        "output": "/scratch/output",
        "msfile": "/scratch/msdir",
    },
    "singularity": {
        "input": "/scratch/input",
        "output": "/scratch/output",
        "msfile": "/scratch/msdir",
    },
}


class StimelaCabParameterError(Exception):
    pass

class StimelaRecipeExecutionError(Exception):
    pass

class PipelineException(Exception):
    """ 
    Encapsulates information about state of pipeline when an
    exception occurs
    """

    def __init__(self, exception, completed, failed, remaining):
        message = ("Exception occurred while running "
                   "pipeline component '%s': %s" % (failed.label, str(exception)))

        super(PipelineException, self).__init__(message)

        self._completed = completed
        self._failed = failed
        self._remaining = remaining

    @property
    def completed(self):
        return self._completed

    @property
    def failed(self):
        return self._failed

    @property
    def remaining(self):
        return self._remaining


class StimelaJob(object):
    def __init__(self, name, recipe, label=None,
                 jtype='docker', cpus=None, memory_limit=None,
                 singularity_dir=None,
                 time_out=-1,
                 log_dir=None):

        self.name = name
        self.recipe = recipe
        self.label = label or '{0}_{1}'.format(name, id(name))
        self.log = recipe.log
        self.active = False
        self.jtype = jtype  # ['docker', 'python', singularity', 'udocker']
        self.job = None
        self.created = False
        self.args = ['--user {}:{}'.format(UID, GID)]
        if cpus:
            self.args.append("--cpus {0:f}".format(cpus))
        if memory_limit:
            self.args.append("--memory {0:s}".format(memory_limit))
        self.time_out = time_out
        self.log_dir = log_dir

    def run_python_job(self):
        function = self.job['function']
        options = self.job['parameters']
        function(**options)
        return 0

    def run_docker_job(self):
        if hasattr(self.job, '_cab'):
            self.job._cab.update(self.job.config,
                                 self.job.parameter_file_name)

        self.created = False
        self.job.create(*self.args)
        self.created = True
        self.job.start()
        return 0

    def run_podman_job(self):
        if hasattr(self.job, '_cab'):
            self.job._cab.update(self.job.config,
                                 self.job.parameter_file_name)

        self.created = False
        self.job.create(*self.args)
        self.created = True
        self.job.start()
        return 0

    def run_singularity_job(self):
        if hasattr(self.job, '_cab'):
            self.job._cab.update(self.job.config,
                                 self.job.parameter_file_name)

        self.created = False
        self.job.start()
        self.created = True
        self.job.run()
        return 0

    def run_udocker_job(self):
        if hasattr(self.job, '_cab'):
            self.job._cab.update(self.job.config,
                                 self.job.parameter_file_name)

        self.created = False
        self.job.create()
        self.created = True
        self.job.run()
        return 0

    def python_job(self, function, parameters=None):
        """
        Run python function

        function    :   Python callable to execute
        name        :   Name of function (if not given, will used function.__name__)
        parameters  :   Parameters to parse to function
        label       :   Function label; for logging purposes
        """

        if not callable(function):
            raise utils.StimelaCabRuntimeError(
                'Object given as function is not callable')

        if self.name is None:
            self.name = function.__name__

        self.job = {
            'function':   function,
            'parameters':   parameters,
        }

        return 0

    def podman_job(self, image, config,
                   input=None, output=None, msdir=None,
                   **kw):
        """
            Run task in podman

        image   :   stimela cab name, e.g. 'cab/simms'
        name    :   This name will be part of the name of the contaier that will 
                    execute the task (now optional)
        config  :   Dictionary of options to parse to the task. This will modify 
                    the parameters in the default parameter file which 
                    can be viewd by running 'stimela cabs -i <cab name>', e.g 'stimela cabs -i simms'
        input   :   input dirctory for cab
        output  :   output directory for cab
        msdir   :   MS directory for cab. Only specify if different from recipe ms_dir


        """

        # check if name has any offending charecters
        offenders = re.findall('\W', self.name)
        if offenders:
            raise StimelaCabParameterError('The cab name \'{:s}\' has some non-alphanumeric characters.'
                                           ' Charecters making up this name must be in [a-z,A-Z,0-9,_]'.format(self.name))

        # Update I/O with values specified on command line
        # TODO (sphe) I think this feature should be removed
        script_context = self.recipe.stimela_context
        input = script_context.get('_STIMELA_INPUT', None) or input
        output = script_context.get('_STIMELA_OUTPUT', None) or output
        msdir = script_context.get('_STIMELA_MSDIR', None) or msdir

        # Get location of template parameters file
        cabpath = self.recipe.stimela_path + \
            "/cargo/cab/{0:s}/".format(image.split("/")[1])
        parameter_file = cabpath+'/parameters.json'

        name = '{0}-{1}{2}'.format(self.name, id(image),
                                   str(time.time()).replace('.', ''))

        _cab = cab.CabDefinition(indir=input, outdir=output,
                                 msdir=msdir, parameter_file=parameter_file)

        cab.IODEST = CONT_IO["podman"]

        cont = podman.Container(image, name,
                                logger=self.log, time_out=self.time_out)

        # Container parameter file will be updated and validated before the container is executed
        cont._cab = _cab
        cont.parameter_file_name = '{0}/{1}.json'.format(
            self.recipe.parameter_file_dir, name)

        # Remove dismissable kw arguments:
        ops_to_pop = []
        for op in config:
            if isinstance(config[op], dismissable):
                ops_to_pop.append(op)
        for op in ops_to_pop:
            arg = config.pop(op)()
            if arg is not None:
                config[op] = arg
        cont.config = config

        # These are standard volumes and
        # environmental variables. These will be
        # always exist in a cab container
        cont.add_volume(self.recipe.stimela_path,
                        '/scratch/stimela', perm='ro')
        cont.add_volume(cont.parameter_file_name,
                        '/scratch/configfile', perm='ro', noverify=True)
        cont.add_volume("{0:s}/cargo/cab/{1:s}/src/".format(
            self.recipe.stimela_path, _cab.task), "/scratch/code", "ro")
        cont.add_volume("{0:s}/cargo/cab/docker_run".format(self.recipe.stimela_path),
                        "/podman_run", perm="ro")
        cont.COMMAND = "/bin/sh -c /podman_run"

        cont.add_environ('CONFIG', '/scratch/configfile'.format(name))

        if msdir:
            md = cab.IODEST["msfile"]
            cont.add_volume(msdir, md)
            cont.add_environ("MSDIR", md)
            # Keep a record of the content of the
            # volume
            dirname, dirs, files = [a for a in next(os.walk(msdir))]
            cont.msdir_content = {
                "volume":   dirname,
                "dirs":   dirs,
                "files":   files,
            }

            self.log.debug(
                'Mounting volume \'{0}\' from local file system to \'{1}\' in the container'.format(msdir, md))

        if input:
            cont.add_volume(input, cab.IODEST["input"], perm='ro')
            cont.add_environ("INPUT", cab.IODEST["input"])
            # Keep a record of the content of the
            # volume
            dirname, dirs, files = [a for a in next(os.walk(input))]
            cont.input_content = {
                "volume":   dirname,
                "dirs":   dirs,
                "files":   files,
            }

            self.log.debug('Mounting volume \'{0}\' from local file system to \'{1}\' in the container'.format(
                input, cab.IODEST["input"]))

        if not os.path.exists(output):
            os.mkdir(output)

        od = cab.IODEST["output"]

        self.log_dir = os.path.abspath(self.log_dir or output)
        log_dir_name = os.path.basename(self.log_dir or output)
        logfile_name = 'log-{0:s}.txt'.format(name.split('-')[0])
        self.logfile = cont.logfile = '{0:s}/{1:s}'.format(
            self.log_dir, logfile_name)

        if not os.path.exists(self.logfile):
            with open(self.logfile, 'w') as std:
                pass

        cont.add_environ("LOGFILE", "/scratch/logfile")
        cont.add_volume(self.logfile, "/scratch/logfile", "rw")
        cont.add_volume(output, od, "rw")
        cont.add_environ("OUTPUT", od)
        self.log.debug(
            'Mounting volume \'{0}\' from local file system to \'{1}\' in the container'.format(output, od))

        cont.image = ":".join([_cab.base, _cab.tag])
        # Added and ready for execution
        self.job = cont

        return 0

    def singularity_job(self, image, config, singularity_image_dir,
                        input=None, output=None, msdir=None,
                        **kw):
        """
            Run task in singularity

        image   :   stimela cab name, e.g. 'cab/simms'
        name    :   This name will be part of the name of the contaier that will 
                    execute the task (now optional)
        config  :   Dictionary of options to parse to the task. This will modify 
                    the parameters in the default parameter file which 
                    can be viewd by running 'stimela cabs -i <cab name>', e.g 'stimela cabs -i simms'
        input   :   input dirctory for cab
        output  :   output directory for cab
        msdir   :   MS directory for cab. Only specify if different from recipe ms_dir


        """

        # check if name has any offending charecters
        offenders = re.findall('\W', self.name)
        if offenders:
            raise StimelaCabParameterError('The cab name \'{:s}\' has some non-alphanumeric characters.'
                                           ' Charecters making up this name must be in [a-z,A-Z,0-9,_]'.format(self.name))

        # Update I/O with values specified on command line
        # TODO (sphe) I think this feature should be removed
        script_context = self.recipe.stimela_context
        input = script_context.get('_STIMELA_INPUT', None) or input
        output = script_context.get('_STIMELA_OUTPUT', None) or output
        msdir = script_context.get('_STIMELA_MSDIR', None) or msdir

        # Get location of template parameters file
        cabpath = self.recipe.stimela_path + \
            "/cargo/cab/{0:s}/".format(image.split("/")[1])
        parameter_file = cabpath+'/parameters.json'

        name = '{0}-{1}{2}'.format(self.name, id(image),
                                   str(time.time()).replace('.', ''))

        _cab = cab.CabDefinition(indir=input, outdir=output,
                                 msdir=msdir, parameter_file=parameter_file)

        cab.IODEST = CONT_IO["singularity"]

        cont = singularity.Container(image, name,
                                     logger=self.log, time_out=self.time_out)

        # Container parameter file will be updated and validated before the container is executed
        cont._cab = _cab
        cont.parameter_file_name = '{0}/{1}.json'.format(
            self.recipe.parameter_file_dir, name)

        # Remove dismissable kw arguments:
        ops_to_pop = []
        for op in config:
            if isinstance(config[op], dismissable):
                ops_to_pop.append(op)
        for op in ops_to_pop:
            arg = config.pop(op)()
            if arg is not None:
                config[op] = arg
        cont.config = config

        # These are standard volumes and
        # environmental variables. These will be
        # always exist in a cab container
        cont.add_volume(self.recipe.stimela_path,
                        '/scratch/stimela', perm='ro')
        cont.add_volume(cont.parameter_file_name,
                        '/scratch/configfile', perm='ro', noverify=True)
        cont.add_volume("{0:s}/cargo/cab/{1:s}/src/".format(
            self.recipe.stimela_path, _cab.task), "/scratch/code", "ro")
        cont.add_volume("{0:s}/cargo/cab/singularity_run".format(self.recipe.stimela_path,
                                                                 _cab.task), "/singularity")

        if msdir:
            md = cab.IODEST["msfile"]
            cont.add_volume(msdir, md)
            # Keep a record of the content of the
            # volume
            dirname, dirs, files = [a for a in next(os.walk(msdir))]
            cont.msdir_content = {
                "volume":   dirname,
                "dirs":   dirs,
                "files":   files,
            }

            self.log.debug(
                'Mounting volume \'{0}\' from local file system to \'{1}\' in the container'.format(msdir, md))

        if input:
            cont.add_volume(input, cab.IODEST["input"], perm='ro')
            # Keep a record of the content of the
            # volume
            dirname, dirs, files = [a for a in next(os.walk(input))]
            cont.input_content = {
                "volume":   dirname,
                "dirs":   dirs,
                "files":   files,
            }

            self.log.debug('Mounting volume \'{0}\' from local file system to \'{1}\' in the container'.format(
                input, cab.IODEST["input"]))

        if not os.path.exists(output):
            os.mkdir(output)

        od = cab.IODEST["output"]

        self.log_dir = os.path.abspath(self.log_dir or output)
        log_dir_name = os.path.basename(self.log_dir or output)
        logfile_name = 'log-{0:s}.txt'.format(name.split('-')[0])
        self.logfile = cont.logfile = '{0:s}/{1:s}'.format(
            self.log_dir, logfile_name)

        if not os.path.exists(self.logfile):
            with open(self.logfile, 'w') as std:
                pass

        cont.add_volume(self.logfile, "/scratch/logfile", "rw")
        cont.add_volume(output, od, "rw")
        self.log.debug(
            'Mounting volume \'{0}\' from local file system to \'{1}\' in the container'.format(output, od))

        simage = _cab.base.replace("/", "_")
        cont.image = '{0:s}/{1:s}_{2:s}.img'.format(
            singularity_image_dir, simage, _cab.tag)
        # Added and ready for execution
        self.job = cont

        return 0

    def udocker_job(self, image, config,
                    input=None, output=None, msdir=None,
                    **kw):
        """
            Run task using udocker

        image   :   stimela cab name, e.g. 'cab/simms'
        name    :   This name will be part of the name of the contaier that will 
                    execute the task (now optional)
        config  :   Dictionary of options to parse to the task. This will modify 
                    the parameters in the default parameter file which 
                    can be viewd by running 'stimela cabs -i <cab name>', e.g 'stimela cabs -i simms'
        input   :   input dirctory for cab
        output  :   output directory for cab
        msdir   :   MS directory for cab. Only specify if different from recipe ms_dir


        """

        # check if name has any offending charecters
        offenders = re.findall('\W', self.name)
        if offenders:
            raise StimelaCabParameterError('The cab name \'{:s}\' has some non-alphanumeric characters.'
                                           ' Charecters making up this name must be in [a-z,A-Z,0-9,_]'.format(self.name))

        # Update I/O with values specified on command line
        # TODO (sphe) I think this feature should be removed
        script_context = self.recipe.stimela_context
        input = script_context.get('_STIMELA_INPUT', None) or input
        output = script_context.get('_STIMELA_OUTPUT', None) or output
        msdir = script_context.get('_STIMELA_MSDIR', None) or msdir

        # Get location of template parameters file
        cabpath = self.recipe.stimela_path + \
            "/cargo/cab/{0:s}/".format(image.split("/")[1])
        parameter_file = cabpath+'/parameters.json'

        name = '{0}-{1}{2}'.format(self.name, id(image),
                                   str(time.time()).replace('.', ''))

        _cab = cab.CabDefinition(indir=input, outdir=output,
                                 msdir=msdir, parameter_file=parameter_file)

        cab.IODEST = CONT_IO["udocker"]

        cont = udocker.Container(image, name,
                                 logger=self.log, time_out=self.time_out)

        cont.add_volume(
            "{0:s}/cargo/cab/docker_run".format(self.recipe.stimela_path), "/udocker_run")
        cont.COMMAND = "/bin/sh -c /udocker_run"

        # Container parameter file will be updated and validated before the container is executed
        cont._cab = _cab
        cont.parameter_file_name = '{0}/{1}.json'.format(
            self.recipe.parameter_file_dir, name)

        # Remove dismissable kw arguments:
        ops_to_pop = []
        for op in config:
            if isinstance(config[op], dismissable):
                ops_to_pop.append(op)
        for op in ops_to_pop:
            arg = config.pop(op)()
            if arg is not None:
                config[op] = arg
        cont.config = config

        # These are standard volumes and
        # environmental variables. These will be
        # always exist in a cab container
        cont.add_volume(self.recipe.stimela_path, '/scratch/stimela')
        cont.add_volume(cont.parameter_file_name,
                        '/scratch/configfile', noverify=True)
        cont.add_volume("{0:s}/cargo/cab/{1:s}/src/".format(
            self.recipe.stimela_path, _cab.task), "/scratch/code")

        cont.add_environ('CONFIG', '/scratch/configfile'.format(name))

        if msdir:
            md = cab.IODEST["msfile"]
            cont.add_volume(msdir, md)
            cont.add_environ("MSDIR", md)
            # Keep a record of the content of the
            # volume
            dirname, dirs, files = [a for a in next(os.walk(msdir))]
            cont.msdir_content = {
                "volume":   dirname,
                "dirs":   dirs,
                "files":   files,
            }

            self.log.debug(
                'Mounting volume \'{0}\' from local file system to \'{1}\' in the container'.format(msdir, md))

        if input:
            cont.add_volume(input, cab.IODEST["input"])
            cont.add_environ("INPUT", cab.IODEST["input"])
            # Keep a record of the content of the
            # volume
            dirname, dirs, files = [a for a in next(os.walk(input))]
            cont.input_content = {
                "volume":   dirname,
                "dirs":   dirs,
                "files":   files,
            }

            self.log.debug('Mounting volume \'{0}\' from local file system to \'{1}\' in the container'.format(
                input, cab.IODEST["input"]))

        if not os.path.exists(output):
            os.mkdir(output)

        od = cab.IODEST["output"]
        cont.WORKDIR = od

        self.log_dir = os.path.abspath(self.log_dir or output)
        log_dir_name = os.path.basename(self.log_dir or output)
        logfile_name = 'log-{0:s}.txt'.format(name.split('-')[0])
        self.logfile = cont.logfile = '{0:s}/{1:s}'.format(
            self.log_dir, logfile_name)

        if not os.path.exists(self.logfile):
            with open(self.logfile, 'w') as std:
                pass
        cont.add_environ("LOGFILE", "/scratch/logfile")
        cont.add_volume(self.logfile, "/scratch/logfile")
        cont.add_volume(output, od)
        cont.add_environ("OUTPUT", od)
        self.log.debug(
            'Mounting volume \'{0}\' from local file system to \'{1}\' in the container'.format(output, od))

        if hasattr(cont._cab, "use_graphics") and cont._cab.use_graphics:
            cont.use_graphics = True
        cont.image = '{0:s}:{1:s}'.format(_cab.base, _cab.tag)
        # Added and ready for execution
        self.job = cont

        return 0

    def docker_job(self, image, config=None,
                   input=None, output=None, msdir=None,
                   shared_memory='1gb', build_label=None,
                   **kw):
        """
        Add a task to a stimela recipe

        image   :   stimela cab name, e.g. 'cab/simms'
        name    :   This name will be part of the name of the contaier that will 
                    execute the task (now optional)
        config  :   Dictionary of options to parse to the task. This will modify 
                    the parameters in the default parameter file which 
                    can be viewd by running 'stimela cabs -i <cab name>', e.g 'stimela cabs -i simms'
        input   :   input dirctory for cab
        output  :   output directory for cab
        msdir   :   MS directory for cab. Only specify if different from recipe ms_dir
        """

        # check if name has any offending charecters
        offenders = re.findall('\W', self.name)
        if offenders:
            raise StimelaCabParameterError('The cab name \'{:s}\' has some non-alphanumeric characters.'
                                           ' Charecters making up this name must be in [a-z,A-Z,0-9,_]'.format(self.name))

        # Update I/O with values specified on command line
        # TODO (sphe) I think this feature should be removed
        script_context = self.recipe.stimela_context
        input = script_context.get('_STIMELA_INPUT', None) or input
        output = script_context.get('_STIMELA_OUTPUT', None) or output
        output = os.path.abspath(output)
        msdir = script_context.get('_STIMELA_MSDIR', None) or msdir
        build_label = script_context.get(
            '_STIMELA_BUILD_LABEL', None) or build_label

        # Get location of template parameters file
        cabs_logger = get_cabs(
            '{0:s}/{1:s}_stimela_logfile.json'.format(stimela.LOG_HOME, build_label))
        try:
            cabpath = cabs_logger['{0:s}_{1:s}'.format(
                build_label, image)]['DIR']
        except KeyError:
            raise StimelaCabParameterError(
                'Cab {} has is uknown to stimela. Was it built?'.format(image))
        parameter_file = cabpath+'/parameters.json'

        name = '{0}-{1}{2}'.format(self.name, id(image),
                                   str(time.time()).replace('.', ''))

        _cab = cab.CabDefinition(indir=input, outdir=output,
                                 msdir=msdir, parameter_file=parameter_file)

        cont = docker.Container(image, name,
                                label=self.label, logger=self.log,
                                shared_memory=shared_memory,
                                log_container=stimela.LOG_FILE,
                                time_out=self.time_out)

        # Container parameter file will be updated and validated before the container is executed
        cont._cab = _cab
        cont.parameter_file_name = '{0}/{1}.json'.format(
            self.recipe.parameter_file_dir, name)

        # Remove dismissable kw arguments:
        ops_to_pop = []
        for op in config:
            if isinstance(config[op], dismissable):
                ops_to_pop.append(op)
        for op in ops_to_pop:
            arg = config.pop(op)()
            if arg is not None:
                config[op] = arg
        cont.config = config

        cont.add_volume(
            "{0:s}/cargo/cab/docker_run".format(self.recipe.stimela_path), "/docker_run", perm="ro")
        cont.COMMAND = "/bin/sh -c /docker_run"
        # These are standard volumes and
        # environmental variables. These will be
        # always exist in a cab container
        cont.add_volume(self.recipe.stimela_path,
                        '/scratch/stimela', perm='ro')
        cont.add_volume(self.recipe.parameter_file_dir, '/configs', perm='ro')
        cont.add_environ('CONFIG', '/configs/{}.json'.format(name))

        cab.IODEST = CONT_IO["docker"]

        if msdir:
            md = cab.IODEST["msfile"]
            cont.add_volume(msdir, md)
            cont.add_environ('MSDIR', md)
            # Keep a record of the content of the
            # volume
            dirname, dirs, files = [a for a in next(os.walk(msdir))]
            cont.msdir_content = {
                "volume":   dirname,
                "dirs":   dirs,
                "files":   files,
            }

            self.log.debug(
                'Mounting volume \'{0}\' from local file system to \'{1}\' in the container'.format(msdir, md))

        if input:
            cont.add_volume(input, cab.IODEST["input"], perm='ro')
            cont.add_environ('INPUT', cab.IODEST["input"])
            # Keep a record of the content of the
            # volume
            dirname, dirs, files = [a for a in next(os.walk(input))]
            cont.input_content = {
                "volume":   dirname,
                "dirs":   dirs,
                "files":   files,
            }

            self.log.debug('Mounting volume \'{0}\' from local file system to \'{1}\' in the container'.format(
                input, cab.IODEST["input"]))

        if not os.path.exists(output):
            os.mkdir(output)

        od = cab.IODEST["output"]
        cont.add_environ('HOME', od)
        cont.add_environ('OUTPUT', od)

        self.log_dir = os.path.abspath(self.log_dir or output)
        log_dir_name = os.path.basename(self.log_dir or output)
        logfile_name = 'log-{0:s}.txt'.format(name.split('-')[0])
        self.logfile = cont.logfile = '{0:s}/{1:s}'.format(
            self.log_dir, logfile_name)
        cont.add_volume(output, od, "rw")

        if not os.path.exists(self.logfile):
            with open(self.logfile, "w") as std:
                pass
        cont.add_volume(
            self.logfile, "{0:s}/logfile".format(self.log_dir), "rw")
        cont.add_environ('LOGFILE',  "{0:}/logfile".format(self.log_dir))
        self.log.debug(
            'Mounting volume \'{0}\' from local file system to \'{1}\' in the container'.format(output, od))

        cont.image = '{0}_{1}'.format(build_label, image)
        # Added and ready for execution
        self.job = cont

        return 0


class Recipe(object):
    def __init__(self, name, data=None,
                 parameter_file_dir=None, ms_dir=None,
                 tag=None, build_label=None, loglevel='INFO',
                 loggername='STIMELA', singularity_image_dir=None, log_dir=None, JOB_TYPE='docker'):
        """
        Deifine and manage a stimela recipe instance.        

        name    :   Name of stimela recipe
        msdir   :   Path of MSs to be used during the execution of the recipe
        tag     :   Use cabs with a specific tag
        parameter_file_dir :   Will store task specific parameter files here
        """

        self.log = logging.getLogger(loggername)
        self.log.setLevel(getattr(logging, loglevel))

        self.JOB_TYPE = JOB_TYPE

        name_ = name.lower().replace(' ', '_')
        self.log_dir = log_dir
        if self.log_dir:
            if not os.path.exists(self.log_dir):
                self.log.info(
                    'The Log directory \'{0:s}\' cannot be found. Will create it'.format(self.log_dir))
                os.makedirs(self.log_dir)

        logfile_name = 'log-{0:s}.txt'.format(name_.split('-')[0])
        self.logfile = '{0}/{1}'.format(self.log_dir or ".", logfile_name)

        # Create file handler which logs even debug
        # messages
        self.resume_file = '.last_{}.json'.format(name_)

        fh = logging.FileHandler(self.logfile, 'w')
        fh.setLevel(logging.DEBUG)
        # Create console handler with a higher log level
        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(getattr(logging, loglevel))
        # Create formatter and add it to the handlers
        formatter = logging.Formatter(
            '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        ch.setFormatter(formatter)
        fh.setFormatter(formatter)
        # Add the handlers to logger

        len(list(filter(lambda x: isinstance(x, logging.StreamHandler),
                        self.log.handlers))) == 0 and self.log.addHandler(ch)
        len(list(filter(lambda x: isinstance(x, logging.FileHandler),
                        self.log.handlers))) == 0 and self.log.addHandler(fh)

        self.stimela_context = inspect.currentframe().f_back.f_globals

        self.stimela_path = os.path.dirname(docker.__file__)

        self.name = name
        self.build_label = build_label or USER
        self.ms_dir = ms_dir
        if not os.path.exists(self.ms_dir):
            self.log.info(
                'MS directory \'{}\' does not exist. Will create it'.format(self.ms_dir))
            os.mkdir(self.ms_dir)
        self.tag = tag
        # create a folder to store config files
        # if it doesn't exist. These config
        # files can be resued to re-run the
        # task
        self.parameter_file_dir = parameter_file_dir or "stimela_parameter_files"
        if not os.path.exists(self.parameter_file_dir):
            self.log.info(
                'Config directory cannot be found. Will create ./{}'.format(self.parameter_file_dir))
            os.mkdir(self.parameter_file_dir)

        self.jobs = []
        self.completed = []
        self.failed = None
        self.remaining = []

        #self.proc_logger = utils.logger.StimelaLogger(stimela.LOG_FILE)
        self.pid = os.getpid()
        #self.proc_logger.log_process(self.pid, self.name)
        # self.proc_logger.write()
        self.singularity_image_dir = singularity_image_dir
        if self.singularity_image_dir:
            self.JOB_TYPE = "singularity"

        self.log.info('---------------------------------')
        self.log.info('Stimela version {0}'.format(stimela.__version__))
        self.log.info('Sphesihle Makhathini <sphemakh@gmail.com>')
        self.log.info('Running: {:s}'.format(self.name))
        self.log.info('---------------------------------')

    def add(self, image, name, config=None,
            input=None, output=None, msdir=None,
            label=None, shared_memory='1gb',
            build_label=None,
            cpus=None, memory_limit=None,
            time_out=-1,
            log_dir=None):

        if self.log_dir:
            if not os.path.exists(self.log_dir):
                self.log.info(
                    'The Log directory \'{0:s}\' cannot be found. Will create it'.format(self.log_dir))
                os.mkdir(log_dir)
        else:
            log_dir = self.log_dir

        job = StimelaJob(name, recipe=self, label=label,
                         cpus=cpus, memory_limit=memory_limit, time_out=time_out,
                         log_dir=self.log_dir or output, jtype=self.JOB_TYPE)

        if callable(image):
            job.jtype = 'function'
            job.python_job(image, parameters=config)
            self.jobs.append(job)
            self.log.info('Adding Python job \'{0}\' to recipe.'.format(name))
        else:
            job.jtype = self.JOB_TYPE
            job_func = getattr(job, "{0:s}_job".format(job.jtype))
            job_func(image=image, config=config,
                     input=input, output=output, msdir=msdir or self.ms_dir,
                     shared_memory=shared_memory, build_label=build_label or self.build_label,
                     singularity_image_dir=self.singularity_image_dir,
                     time_out=time_out)

            self.log.info('Adding cab \'{0}\' to recipe. The container will be named \'{1}\''.format(
                job.job.image, name))
            self.jobs.append(job)

        return 0

    def log2recipe(self, job, recipe, num, status):

        if job.jtype in ['docker', 'singularity', 'udocker', 'podman']:
            cont = job.job
            step = {
                "name":   cont.name,
                "number":   num,
                "cab":   cont.image,
                "volumes":   cont.volumes,
                "environs":   getattr(cont, "environs", None),
                "shared_memory":   getattr(cont, "shared_memory", None),
                "input_content":   cont.input_content,
                "msdir_content":   cont.msdir_content,
                "label":   getattr(cont, "label", ""),
                "logfile":   cont.logfile,
                "status":   status,
                "jtype":   'docker',
            }
        else:
            step = {
                "name":   job.name,
                "number":   num,
                "label":   job.label,
                "status":   status,
                "function":   job.job['function'].__name__,
                "jtype":   'function',
                "parameters":   job.job['parameters'],
            }

        recipe['steps'].append(step)

        return 0

    def run(self, steps=None, resume=False, redo=None):
        """
        Run a Stimela recipe. 

        steps   :   recipe steps to run
        resume  :   resume recipe from last run
        redo    :   Re-run an old recipe from a .last file
        """

        recipe = {
            "name":   self.name,
            "steps":   []
        }
        start_at = 0

        if redo:
            recipe = utils.readJson(redo)
            self.log.info('Rerunning recipe {0} from {1}'.format(
                recipe['name'], redo))
            self.log.info('Recreating recipe instance..')
            self.jobs = []
            for step in recipe['steps']:

                #        add I/O folders to the json file
                #        add a string describing the contents of these folders
                #        The user has to ensure that these folders exist, and have the required content
                if step['jtype'] == 'docker':
                    self.log.info('Adding job \'{0}\' to recipe. The container will be named \'{1}\''.format(
                        step['cab'], step['name']))
                    cont = docker.Container(step['cab'], step['name'],
                                            label=step['label'], logger=self.log,
                                            shared_memory=step['shared_memory'])

                    self.log.debug('Adding volumes {0} and environmental variables {1}'.format(
                        step['volumes'], step['environs']))
                    cont.volumes = step['volumes']
                    cont.environs = step['environs']
                    cont.shared_memory = step['shared_memory']
                    cont.input_content = step['input_content']
                    cont.msdir_content = step['msdir_content']
                    cont.logfile = step['logfile']
                    job = StimelaJob(
                        step['name'], recipe=self, label=step['label'])
                    job.job = cont
                    job.jtype = 'docker'

                elif step['jtype'] == 'function':
                    name = step['name']
                    func = inspect.currentframe(
                    ).f_back.f_locals[step['function']]
                    job = StimelaJob(name, recipe=self, label=step['label'])
                    job.python_job(func, step['parameters'])
                    job.jtype = 'function'

                self.jobs.append(job)

        elif resume:
            self.log.info("Resuming recipe from last run.")
            try:
                recipe = utils.readJson(self.resume_file)
            except IOError:
                raise StimelaRecipeExecutionError(
                    "Cannot resume pipeline, resume file '{}' not found".format(self.resume_file))

            steps_ = recipe.pop('steps')
            recipe['steps'] = []
            _steps = []
            for step in steps_:
                if step['status'] == 'completed':
                    recipe['steps'].append(step)
                    continue

                label = step['label']
                number = step['number']

                # Check if the recipe flow has changed
                if label == self.jobs[number-1].label:
                    self.log.info(
                        'recipe step \'{0}\' is fit for re-execution. Label = {1}'.format(number, label))
                    _steps.append(number)
                else:
                    raise StimelaRecipeExecutionError(
                        'Recipe flow, or task scheduling has changed. Cannot resume recipe. Label = {0}'.format(label))

            # Check whether there are steps to resume
            if len(_steps) == 0:
                self.log.info(
                    'All the steps were completed. No steps to resume')
                sys.exit(0)
            steps = _steps

        if getattr(steps, '__iter__', False):
            _steps = []
            if isinstance(steps[0], str):
                labels = [job.label.split('::')[0] for job in self.jobs]

                for step in steps:
                    try:
                        _steps.append(labels.index(step)+1)
                    except ValueError:
                        raise StimelaCabParameterError(
                            'Recipe label ID [{0}] doesn\'t exist'.format(step))
                steps = _steps
        else:
            steps = range(1, len(self.jobs)+1)

        jobs = [(step, self.jobs[step-1]) for step in steps]

        for i, (step, job) in enumerate(jobs):

            self.log.info('Running job {}'.format(job.name))
            self.log.info('STEP {0} :: {1}'.format(i+1, job.label))
            self.active = job
            try:
                if job.jtype == 'function':
                    job.run_python_job()
                elif job.jtype in ['docker', 'singularity', 'udocker', 'podman']:
                    with open(job.job.logfile, 'a') as astd:
                        astd.write('\n-----------------------------------\n')
                        astd.write(
                            'Stimela version     : {}\n'.format(version))
                        astd.write(
                            'Cab name            : {}\n'.format(job.job.image))
                        astd.write('-------------------------------------\n')

                    run_job = getattr(job, "run_{0:s}_job".format(job.jtype))
                    run_job()

                self.log2recipe(job, recipe, step, 'completed')

            except (utils.StimelaCabRuntimeError,
                    StimelaRecipeExecutionError,
                    StimelaCabParameterError) as e:
                self.completed = [jb[1] for jb in jobs[:i]]
                self.remaining = [jb[1] for jb in jobs[i+1:]]
                self.failed = job

                self.log.info(
                    'Recipe execution failed while running job {}'.format(job.name))
                self.log.info('Completed jobs : {}'.format(
                    [c.name for c in self.completed]))
                self.log.info('Remaining jobs : {}'.format(
                    [c.name for c in self.remaining]))

                self.log2recipe(job, recipe, step, 'failed')
                for step, jb in jobs[i+1:]:
                    self.log.info(
                        'Logging remaining task: {}'.format(jb.label))
                    self.log2recipe(jb, recipe, step, 'remaining')

                self.log.info(
                    'Saving pipeline information in {}'.format(self.resume_file))
                utils.writeJson(self.resume_file, recipe)

                pe = PipelineException(e, self.completed, job, self.remaining)
                raise_(pe, None, sys.exc_info()[2])
            except:
                import traceback
                traceback.print_exc()
                raise RuntimeError(
                    "An unhandled exception has occured. This is a bug, please report")

            finally:
                if job.jtype == 'docker' and job.created:
                    job.job.stop()
                    job.job.remove()
                elif job.jtype == 'singularity' and job.created:
                    job.job.stop()

        self.log.info(
            'Saving pipeline information in {}'.format(self.resume_file))
        utils.writeJson(self.resume_file, recipe)

        self.log.info('Recipe executed successfully')

        return 0



import os
import sys
import tempfile

sys.path.append('/scratch/stimela')

utils = __import__('utils')


CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
MSDIR = os.environ["MSDIR"]

cab = utils.readJson(CONFIG)
args = []
msname = None

custom_script = "print(\"Nothing has been done\")"

for param in cab['parameters']:
    name = param['name']
    value = param['value']
    if name == "script":
        custom_script = value
        continue

with tempfile.NamedTemporaryFile(suffix=".py") as tfile:
    tfile.write(custom_script)
    tfile.flush()
    utils.xrun(cab['binary'], [tfile.name])



import os
import sys

sys.path.append("/scratch/stimela")

utils = __import__('utils')

CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
OUTPUT = os.environ["OUTPUT"]
MSDIR = os.environ["MSDIR"]

cab = utils.readJson(CONFIG)
args = []
for param in cab['parameters']:
    name = param['name']
    value = param['value']

    if name == 'ms':
        mslist = ' '.join(value)
        continue
    if value in [None, False]:
        continue
    if value is True:
        value = ""
    args += ['{0}{1} {2}'.format(cab['prefix'], name, value)]

utils.xrun("mask_ms.py", args + [mslist])



import os
import sys
import yaml
import sharpener

sys.path.append('/scratch/stimela')

utils = __import__('utils')


CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
MSDIR = os.environ["MSDIR"]
OUTPUT = os.environ["OUTPUT"]

cab = utils.readJson(CONFIG)
args = []
msname = None

pkg_path = os.path.dirname(os.path.realpath(sharpener.__file__))
sharpener_file = '{:s}/sharpener_default.yml'.format(pkg_path)

with open(sharpener_file) as f:
    list_doc = yaml.load(f)

list_doc['general']['workdir'] = '{:s}/'.format(OUTPUT)

for param in cab['parameters']:
    name = param['name']
    value = param['value']

    if value is None:
        continue

    if name in ['contname', 'cubename']:
        list_doc['general'][name] = value.split('/')[-1]
        continue
    if name in ['catalog_file']:
        list_doc['source_catalog'][name] = value.split('/')[-1]
        continue

    for key, val in list_doc.items():
        if type(val) == dict:
            for k1, v1 in val.items():
                if 'enable' in name:
                    if key in name:
                        list_doc[key]['enable'] = value
                elif k1 == name:
                    list_doc[key][k1] = value
        else:
            if key == name:
                list_doc[key] = value

edited_file = 'sharpener_default.yml'
with open(edited_file, "w") as f:
    yaml.dump(list_doc, f)

utils.xrun('run_sharpener -c ', [edited_file])



import sys
import os
from MSUtils import msutils
import MSUtils.ClassESW as esw
import inspect
from MSUtils.imp_plotter import gain_plotter

sys.path.append("/scratch/stimela")

utils = __import__('utils')

CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
OUTPUT = os.environ["OUTPUT"]
MSDIR = os.environ["MSDIR"]

cab = utils.readJson(CONFIG)

args = {}
for param in cab['parameters']:
    name = param['name']
    value = param['value']

    if name == "command":
        function = value
        continue

    args[name] = value

if function == 'sumcols':
    args['outcol'] = args.pop('colname')

if function == "estimate_weights":
    msnoise = esw.MSNoise(args['msname'])
    if isinstance(args['stats_data'], str) and args['stats_data'].find('use_package_meerkat_spec') >= 0:
        args['stats_data'] = esw.MEERKAT_SEFD

    # Calculate noise/weights from spec
    noise, weights = msnoise.estimate_weights(stats_data=args['stats_data'],
                                              smooth=args['smooth'],
                                              fit_order=args['fit_order'],
                                              plot_stats=args.get('plot_stats', None))

    if args['write_to_ms']:
        msnoise.write_toms(noise, columns=args['noise_columns'])
        msnoise.write_toms(weights, columns=args['weight_columns'])
    sys.exit(0)

if function == "plot_gains":
    tab = args['ctable']
    tabtype = args['tabtype']
    dpi = args['plot_dpi']
    scale = args['subplot_scale']
    outfile = args['plot_file']
    gain_plotter(tab, tabtype, outfile, scale, dpi)
    sys.exit(0)


run_func = getattr(msutils, function, None)
if run_func is None:
    raise RuntimeError("Function '{}' is not part of MSUtils".format(function))

# Reove default parameters that are not part of this particular function
func_args = inspect.getargspec(run_func)[0]
for arg in args.keys():
    if arg not in func_args:
        args.pop(arg, None)

run_func(**args)



import os
import sys
import logging
import Crasa.Crasa as crasa

sys.path.append("/scratch/stimela")

utils = __import__('utils')

CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
OUTPUT = os.environ["OUTPUT"]
MSDIR = os.environ["MSDIR"]

cab = utils.readJson(CONFIG)

makemask_args = {}
immath_args = {}
for param in cab['parameters']:
    name = param['name']
    value = param['value']

    if value is None:
        continue

    if name in ['threshold', 'inpimage', 'output']:
        if name in ['threshold']:
            im_value = ' iif( IM0 >=%s, IM0, 0.0) ' % value
            im_name = 'expr'
        if name in ['output']:
            im_value = '%s_thresh' % value
            im_name = 'outfile'
        if name in ['inpimage']:
            im_value = value
            im_name = 'imagename'
        immath_args[im_name] = im_value

    if name in ['mode', 'inpimage', 'inpmask', 'output', 'overwrite']:
        makemask_args[name] = value

if 'expr' in immath_args:
    task = crasa.CasaTask("immath", **immath_args)
    task.run()

if 'inpmask' not in makemask_args:
    makemask_args['inpmask'] = immath_args['outfile']

task = crasa.CasaTask(cab["binary"], **makemask_args)
task.run()



import os
import sys

sys.path.append("/scratch/stimela")

utils = __import__('utils')

CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
OUTPUT = os.environ["OUTPUT"]
MSDIR = os.environ["MSDIR"]

cab = utils.readJson(CONFIG)

args = []
inimage = None
outimage = None
stack = False
unstack = False
axis = None
chunk = 1

for param in cab['parameters']:
    value = param['value']
    name = param['name']

    if value in [None, False]:
        continue

    if name == 'image':
        inimage = ' '.join(value)
        continue
    elif name == 'output':
        outimage = value
        continue
    elif name == 'stack':
        stack = True
        continue
    elif name == 'unstack':
        unstack = True
        continue
    elif name == 'unstack-chunk':
        chunk = value
        continue
    elif name == 'fits-axis':
        axis = value
        continue

    elif value is True:
        value = ""

    args.append('{0}{1} {2}'.format(cab['prefix'], name, value))

if stack and axis:
    args.append('{0}stack {1}:{2}'.format(cab['prefix'], outimage, axis))
    outimage = None
elif unstack and axis:
    args.append('{0}unstack {1}:{2}:{3}'.format(
        cab['prefix'], outimage, axis, chunk))
    outimage = None
else:
    outimage = '{0}output {1}'.format(cab['prefix'], outimage)
utils.xrun(cab['binary'], args+[inimage, outimage or ""])



import sys
import os
from sunblocker.sunblocker import Sunblocker
import inspect

sys.path.append("/scratch/stimela")

utils = __import__('utils')

CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
OUTPUT = os.environ["OUTPUT"]
MSDIR = os.environ["MSDIR"]

cab = utils.readJson(CONFIG)

args = {}
for param in cab['parameters']:
    name = param['name']
    value = param['value']

    if name == "command":
        function = value
        continue
    if value is None:
        continue

    args[name] = value

args['showdir'] = OUTPUT
run_func = getattr(Sunblocker(), function, None)
if run_func is None:
    raise RuntimeError(
        "Function '{}' is not part of Sunblocker()".format(function))

# Reove default parameters that are not part of this particular function
sys.stdout.write(repr(args))
func_args = inspect.getargspec(run_func)[0]
for arg in args.keys():
    if arg not in func_args:
        args.pop(arg, None)

run_func(**args)



from multiprocessing import Process, Manager, Lock
import subprocess
import signal
import os
import sys
import logging
import json
import yaml
import codecs
import time
import tempfile
import inspect
import warnings
import re
import math
from threading import Thread
import unicodedata
import hashlib
#from fcntl import fcntl, F_GETFL, F_SETFL
#from os import O_NONBLOCK, read

DEBUG = False
INTERRUPT_TIME = 1.0  # seconds -- do not want to constantly interrupt the child process


class StimelaCabRuntimeError(RuntimeError):
    pass


CPUS = 1


def _logger(level=0, logfile=None):

    if logfile and not logging.getLogger("STIMELA"):
        logging.basicConfig(filename=logfile)
    elif not logging.getLogger("STIMELA"):
        logging.basicConfig()

    LOGL = {"0": "INFO",
            "1": "DEBUG",
            "2": "ERROR",
            "3": "CRITICAL"}

    log = logging.getLogger("STIMELA")
    log.setLevel(eval("logging."+LOGL[str(level)]))

    return log


def assign(key, value):
    frame = inspect.currentframe().f_back
    frame.f_globals[key] = value


def xrun(command, options, log=None, _log_container_as_started=False, logfile=None, timeout=-1, kill_callback=None):
    """
        Run something on command line.

        Example: _run("ls", ["-lrt", "../"])
    """

    cmd = " ".join([command] + list(map(str, options)))

    def _print_info(msg):
        if msg is None:
            return
        if log:
            log.info(msg)
        else:
            print(msg)

    def _print_warn(msg):
        if msg is None:
            return
        if log:
            log.warn(msg)
        else:
            print(msg)

    _print_info(u"Running: {0:s}".format(cmd))

    sys.stdout.flush()
    starttime = time.time()
    process = p = None
    try:
        foutname = os.path.join("/tmp", "stimela_output_{0:s}_{1:f}".format(
            hashlib.md5(cmd.encode('utf-8')).hexdigest(), starttime))
        with open(foutname, "w+") as fout:
            p = process = subprocess.Popen(cmd,
                                           stderr=fout,
                                           stdout=fout,
                                           shell=True)
            kill_callback = kill_callback or p.kill

            def clock_killer(p):
                while process.poll() is None and (timeout >= 0):
                    currenttime = time.time()
                    if (currenttime - starttime < timeout):
                        DEBUG and _print_warn(u"Clock Reaper: has been running for {0:f}, must finish in {1:f}".format(
                            currenttime - starttime, timeout))
                    else:
                        _print_warn(
                            u"Clock Reaper: Timeout reached for '{0:s}'... sending the KILL signal".format(cmd))
                        kill_callback()
                    time.sleep(INTERRUPT_TIME)

            Thread(target=clock_killer, args=tuple([p])).start()

            while (process.poll() is None):
                currenttime = time.time()
                DEBUG and _print_info(
                    u"God mode on: has been running for {0:f}".format(currenttime - starttime))
                # this is probably not ideal as it interrupts the process every few seconds,
                time.sleep(INTERRUPT_TIME)
                # check whether there is an alternative with a callback
            assert hasattr(
                process, "returncode"), "No returncode after termination!"
        with open(foutname, "r") as fout:
            _print_info(fout.read())
    finally:
        if (process is not None) and process.returncode:
            raise StimelaCabRuntimeError(
                '%s: returns errr code %d' % (command, process.returncode))


def readJson(conf):
    with open(conf) as _std:
        jdict = yaml.safe_load(_std)
        return jdict


def writeJson(config, dictionary):
    with codecs.open(config, 'w', 'utf8') as std:
        std.write(json.dumps(dictionary, ensure_ascii=False))


def get_Dockerfile_base_image(image):

    if os.path.isfile(image):
        dockerfile = image
    else:
        dockerfile = "{:s}/Dockerfile".format(image)

    with open(dockerfile, "r") as std:
        _from = ""
        for line in std.readlines():
            if line.startswith("FROM"):
                _from = line

    return _from


def change_Dockerfile_base_image(path, _from, label, destdir="."):
    if os.path.isfile(path):
        dockerfile = path
        dirname = os.path.dirname(path)
    else:
        dockerfile = "{:s}/Dockerfile".format(path)
        dirname = path

    with open(dockerfile, "r") as std:
        lines = std.readlines()
        for line in lines:
            if line.startswith("FROM"):
                lines.remove(line)

    temp_dir = tempfile.mkdtemp(
        prefix="tmp-stimela-{:s}-".format(label), dir=destdir)
    xrun(
        "cp", ["-r", "{:s}/Dockerfile {:s}/src".format(dirname, dirname), temp_dir])

    dockerfile = "{:s}/Dockerfile".format(temp_dir)

    with open(dockerfile, "w") as std:
        std.write("{:s}\n".format(_from))

        for line in lines:
            std.write(line)

    return temp_dir, dockerfile


def get_base_images(logfile, index=1):

    with open(logfile) as std:
        string = std.read()

    separator = "[================================DONE==========================]"

    log = string.split(separator)[index-1]

    images = []

    for line in log.split("\n"):
        if line.find("<=BASE_IMAGE=>") > 0:
            tmp = line.split("<=BASE_IMAGE=>")[-1]
            image, base = tmp.split("=")
            images.append((image.strip(), base))

    return images


def icasa(taskname, mult=None, clearstart=False, loadthese=[], **kw0):
    """ 
      runs a CASA task given a list of options.
      A given task can be run multiple times with a different options, 
      in this case the options must be parsed as a list/tuple of dictionaries via mult, e.g 
      icasa('exportfits',mult=[{'imagename':'img1.image','fitsimage':'image1.fits},{'imagename':'img2.image','fitsimage':'image2.fits}]). 
      Options you want be common between the multiple commands should be specified as key word args.
    """

    # create temp directory from which to run casapy
    td = tempfile.mkdtemp(dir='.')
    # we want get back to the working directory once casapy is launched
    cdir = os.path.realpath('.')

    # load modules in loadthese
    _load = ""
    if "os" not in loadthese or "import os" not in loadthese:
        loadthese.append("os")

    if loadthese:
        exclude = filter(lambda line: line.startswith("import")
                         or line.startswith("from"), loadthese)
        for line in loadthese:
            if line not in exclude:
                line = "import %s" % line
            _load += "%s\n" % line

    if mult:
        if isinstance(mult, (tuple, list)):
            for opts in mult:
                opts.update(kw0)
        else:
            mult.upadte(kw0)
            mult = [mult]
    else:
        mult = [kw0]

    run_cmd = """ """
    for kw in mult:
        task_cmds = []
        for key, val in kw.items():
            if isinstance(val, (str, unicode)):
                val = '"%s"' % val
            task_cmds .append('%s=%s' % (key, val))

        task_cmds = ", ".join(task_cmds)
        run_cmd += """ 
%s
os.chdir('%s')
%s
%s(%s)
""" % (_load, cdir, "clearstart()" if clearstart else "", taskname, task_cmds)

    tf = tempfile.NamedTemporaryFile(suffix='.py')
    tf.write(run_cmd)
    tf.flush()
    t0 = time.time()
    # all logging information will be in the pyxis log files
    print("Running {}".format(run_cmd))
    xrun("cd", [td, "&& casa --nologger --log2term --nologfile -c", tf.name])

    # log taskname.last
    task_last = '%s.last' % taskname
    if os.path.exists(task_last):
        with open(task_last, 'r') as last:
            print('%s.last is: \n %s' % (taskname, last.read()))

    # remove temp directory. This also gets rid of the casa log files; so long suckers!
    xrun("rm", ["-fr ", td, task_last])
    tf.close()


def stack_fits(fitslist, outname, axis=0, ctype=None, keep_old=False, fits=False):
    """ Stack a list of fits files along a given axiis.

       fitslist: list of fits file to combine
       outname: output file name
       axis: axis along which to combine the files
       fits: If True will axis FITS ordering axes
       ctype: Axis label in the fits header (if given, axis will be ignored)
       keep_old: Keep component files after combining?
    """
    import numpy
    try:
        import pyfits
    except ImportError:
        warnings.warn(
            "Could not find pyfits on this system. FITS files will not be stacked")
        sys.exit(0)

    hdu = pyfits.open(fitslist[0])[0]
    hdr = hdu.header
    naxis = hdr['NAXIS']

    # find axis via CTYPE key
    if ctype is not None:
        for i in range(1, naxis+1):
            if hdr['CTYPE%d' % i].lower().startswith(ctype.lower()):
                axis = naxis - i  # fits to numpy convention
    elif fits:
        axis = naxis - axis

    fits_ind = abs(axis-naxis)
    crval = hdr['CRVAL%d' % fits_ind]

    imslice = [slice(None)]*naxis
    _sorted = sorted([pyfits.open(fits) for fits in fitslist],
                     key=lambda a: a[0].header['CRVAL%d' % (naxis-axis)])

    # define structure of new FITS file
    nn = [hd[0].header['NAXIS%d' % (naxis-axis)] for hd in _sorted]
    shape = list(hdu.data.shape)
    shape[axis] = sum(nn)
    data = numpy.zeros(shape, dtype=float)

    for i, hdu0 in enumerate(_sorted):
        h = hdu0[0].header
        d = hdu0[0].data
        imslice[axis] = range(sum(nn[:i]), sum(nn[:i+1]))
        data[imslice] = d
        if crval > h['CRVAL%d' % fits_ind]:
            crval = h['CRVAL%d' % fits_ind]

    # update header
    hdr['CRVAL%d' % fits_ind] = crval
    hdr['CRPIX%d' % fits_ind] = 1

    pyfits.writeto(outname, data, hdr, clobber=True)
    print("Successfully stacked images. Output image is %s" % outname)

    # remove old files
    if not keep_old:
        for fits in fitslist:
            os.system('rm -f %s' % fits)


def substitute_globals(string, globs=None):
    sub = set(re.findall('\{(.*?)\}', string))
    globs = globs or inspect.currentframe().f_back.f_globals
    if sub:
        for item in map(str, sub):
            string = string.replace("${%s}" % item, globs[item])
        return string
    else:
        return False


def get_imslice(ndim):
    imslice = []
    for i in xrange(ndim):
        if i < ndim-2:
            imslice.append(0)
        else:
            imslice.append(slice(None))

    return imslice


def addcol(msname, colname=None, shape=None,
           data_desc_type='array', valuetype=None, init_with=0, **kw):
    """ add column to MS 
        msanme : MS to add colmn to
        colname : column name
        shape : shape
        valuetype : data type 
        data_desc_type : 'scalar' for scalar elements and array for 'array' elements
        init_with : value to initialise the column with 
    """
    import numpy
    import pyrap.tables
    tab = pyrap.tables.table(msname, readonly=False)

    try:
        tab.getcol(colname)
        print('Column already exists')

    except RuntimeError:
        print('Attempting to add %s column to %s' % (colname, msname))

        from pyrap.tables import maketabdesc
        valuetype = valuetype or 'complex'

        if shape is None:
            dshape = list(tab.getcol('DATA').shape)
            shape = dshape[1:]

        if data_desc_type == 'array':
            from pyrap.tables import makearrcoldesc
            # God forbid this (or the TIME) column doesn't exist
            coldmi = tab.getdminfo('DATA')
            coldmi['NAME'] = colname.lower()
            tab.addcols(maketabdesc(makearrcoldesc(
                colname, init_with, shape=shape, valuetype=valuetype)), coldmi)

        elif data_desc_type == 'scalar':
            from pyrap.tables import makescacoldesc
            coldmi = tab.getdminfo('TIME')
            coldmi['NAME'] = colname.lower()
            tab.addcols(maketabdesc(makescacoldesc(
                colname, init_with, valuetype=valuetype)), coldmi)

        print('Column added successfuly.')

        if init_with:
            nrows = dshape[0]

            rowchunk = nrows//10 if nrows > 1000 else nrows
            for row0 in range(0, nrows, rowchunk):
                nr = min(rowchunk, nrows-row0)
                dshape[0] = nr
                tab.putcol(colname, numpy.ones(
                    dshape, dtype=valuetype)*init_with, row0, nr)

    tab.close()


def sumcols(msname, col1=None, col2=None, outcol=None, cols=None, suntract=False):
    """ add col1 to col2, or sum columns in 'cols' list.
        If subtract, subtract col2 from col1
    """
    from pyrap.tables import table

    tab = table(msname, readonly=False)
    if cols:
        data = 0
        for col in cols:
            data += tab.getcol(col)
    else:
        if subtract:
            data = tab.getcol(col1) - tab.getcol(col2)
        else:
            data = tab.getcol(col1) + tab.getcol(col2)

    rowchunk = nrows//10 if nrows > 1000 else nrows
    for row0 in range(0, nrows, rowchunk):
        nr = min(rowchunk, nrows-row0)
        tab.putcol(outcol, data[row0:row0+nr], row0, nr)

    tab.close()


def copycol(msname, fromcol, tocol):
    from pyrap.tables import table

    tab = table(msname, readonly=False)
    data = tab.getcol(fromcol)
    if tocol not in tab.colnames():
        addcol(msname, tocol)

    nrows = tab.nrows()
    rowchunk = nrows//10 if nrows > 5000 else nrows
    for row0 in range(0, nrows, rowchunk):
        nr = min(rowchunk, nrows-row0)
        tab.putcol(tocol, data[row0:row0+nr], row0, nr)

    tab.close()


def cab_dict_update(dictionary, key=None, value=None, options=None):
    if options is None:
        options = {key: value}
    for key, value in options.items():
        dictionary[key] = dictionary.pop(key, None) or value
    return dictionary


def compute_vis_noise(msname, sefd, spw_id=0):
    """Computes nominal per-visibility noise"""
    from pyrap.tables import table

    tab = table(msname)
    spwtab = table(msname + "/SPECTRAL_WINDOW")

    freq0 = spwtab.getcol("CHAN_FREQ")[spw_id, 0]
    wavelength = 300e+6/freq0
    bw = spwtab.getcol("CHAN_WIDTH")[spw_id, 0]
    dt = tab.getcol("EXPOSURE", 0, 1)[0]
    dtf = (tab.getcol("TIME", tab.nrows()-1, 1)-tab.getcol("TIME", 0, 1))[0]

    # close tables properly, else the calls below will hang waiting for a lock...
    tab.close()
    spwtab.close()

    print(">>> %s freq %.2f MHz (lambda=%.2fm), bandwidth %.2g kHz, %.2fs integrations, %.2fh synthesis" % (
        msname, freq0*1e-6, wavelength, bw*1e-3, dt, dtf/3600))
    noise = sefd/math.sqrt(abs(2*bw*dt))
    print(">>> SEFD of %.2f Jy gives per-visibility noise of %.2f mJy" %
          (sefd, noise*1000))

    return noise



import os
import sys
import time
from optparse import OptionParser
import pylab as plt
from pyrap.tables import table
import numpy as np
import matplotlib
matplotlib.use('Agg')

matplotlib.rcParams.update({'font.size': 22})


def info(string):
    t = "%d/%d/%d %d:%d:%d" % (time.localtime()[:6])
    print "%s ##INFO: %s" % (t, string)


def abort(string):
    t = "%d/%d/%d %d:%d:%d" % (time.localtime()[:6])
    print "%s ##ABORT: %s" % (t, string)
    sys.exit()


def taper(msname, wc=None, res=None, freq=None, savefig=None):
    tab = table(msname, readonly=False)

    # using butterworth function to create weights
    def bt(w, wc, ws, n): return np.sqrt(1./(1 + (wc/(w-ws))**(2.0*n)))
    sign = -1.0

    if len(res or [0]) > 1 or len(wc or [0]) > 1:
        if wc is None:
            res = np.array(res)
            res = (res/3600.) * np.pi/180.  # Convert resolution to radians
            wc = 1.22*((2.998e8/freq)/res) / \
                2.  # Zen master of the universe (Oleg) says divide by 2
            wc = wc[np.argsort(wc)]
        ws = (wc[1] - wc[0])/2. + wc[0]
        wc = (wc[1] - wc[0])/2.
    else:

        ws, sign = 0, 1.0

        if wc is None:
            try:
                res = res[0]
                res = (res/3600.) * np.pi/180  # Convert resolution to radians
            except TypeError:
                abort("Either of res or wc has to be specified")
            wc = 1.22*((2.998e8/freq)/res) / \
                2.  # Zen master of the universe (Oleg) says divide by 2
        else:
            wc = wc[0]

    uvw = tab.getcol("UVW")
    uvdist = np.sqrt(uvw[:, 0]**2 + uvw[:, 1]**2)
    uvdist[np.where(uvdist == 0)] = 1e-3  # avoid division by zero
    info("Calculating taper weights")
    weight_I = bt(uvdist, wc, ws, sign*20)

    if savefig:
        plt.clf()
        plt.plot(uvdist/1e3, weight_I, ".", ms=0.5)
        plt.xlabel('|uv| [km]')
        plt.ylabel('uv-weight')
        plt.ylim(0, 1.05)
        plt.grid()
        plt.xscale("log")
        plt.savefig(savefig)

    # use same weights for all four stkes parameters [I,Q,U,V]
    weight = np.array([weight_I, weight_I, weight_I, weight_I]).T
    nrows = len(uvw)
    rowchunk = nrows/10
    info("Inserting weights in measurement set")

    for row0 in range(0, nrows, rowchunk):
        nr = min(rowchunk, nrows-row0)
        info("Adding weights to (rows %d to %d)" % (row0, row0+nr-1))
        # update MS with new weights
        tab.putcol("WEIGHT", weight[row0:(row0+nr)], row0, nr)

    info("Done!")
    tab.close()


def reset_weights(msname):
    tab = table(msname, readonly=False)

    weights = tab.getcol("WEIGHT")
    weights[...] = 1
    nrows = len(weights)
    rowchunk = nrows/10
    info("Reseting weights in \'WEIGHT\' column in MS")

    for row0 in range(0, nrows, rowchunk):
        nr = min(rowchunk, nrows-row0)
        info("Reseting (rows %d to %d)" % (row0, row0+nr-1))
        # update MS with new weights
        tab.putcol("WEIGHT", weights[row0:(row0+nr)], row0, nr)

    tab.close()
    info("Done!")



import os
import sys
import inspect
import pkg_resources

try:
    __version__ = pkg_resources.require("stimela")[0].version
except pkg_resources.DistributionNotFound:
    __version__ = "dev"

# Get to know user
USER = os.environ["USER"]
UID = os.getuid()
GID = os.getgid()

root = os.path.dirname(__file__)

CAB_PATH = os.path.join(root, "cargo/cab")
BASE_PATH = os.path.join(root, "cargo/base")

# Set up logging infrastructure
LOG_HOME = os.path.expanduser("~/.stimela")
# This is is the default log file. It logs stimela images, containers and processes
LOG_FILE = "{0:s}/stimela_logfile.json".format(LOG_HOME)

GLOBALS = {'foo': 'bar'}
del GLOBALS['foo']

def register_globals():
    frame = inspect.currentframe().f_back
    frame.f_globals.update(GLOBALS)

# Get base images
# All base images must be on dockerhub
BASE = os.listdir(BASE_PATH)
CAB = list()

for item in os.listdir(CAB_PATH):
    try:
        # These files must exist for a cab image to be valid
        ls_cabdir = os.listdir('{0}/{1}'.format(CAB_PATH, item))
        dockerfile = 'Dockerfile' in ls_cabdir
        paramfile = 'parameters.json' in ls_cabdir
        srcdir = 'src' in ls_cabdir
    except OSError:
        continue
    if dockerfile and paramfile and srcdir:
        CAB.append(item)

from stimela.recipe import Recipe



import os
import sys

sys.path.append('/scratch/stimela')

utils = __import__('utils')


CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
OUTPUT = os.environ["OUTPUT"]
MSDIR = os.environ["MSDIR"]

cab = utils.readJson(CONFIG)
args = []
for param in cab['parameters']:
    name = param['name']
    value = param['value']

    if value in [False, None]:
        continue

    if name in 'restoring-beam scale'.split() and hasattr(value, '__iter__'):
        value = ','.join(value)

    if value is True:
        value = ""
        if name == 'f':
            args.append('-f')
            continue

    # Positional arguments
    if name == 'input-image':
        inim = value
        continue

    elif name == 'input-skymodel':
        inlsm = value
        continue

    elif name == 'output-image':
        outim = value
        continue

    args.append('{0}{1} {2}'.format(cab['prefix'], name, value))

utils.xrun(cab['binary'], args+[inim, inlsm, outim])



import os
import argparse
from argparse import ArgumentParser
import textwrap as _textwrap
import signal
import stimela
from stimela import docker, singularity, udocker, podman, utils
from stimela.utils import logger

BASE = stimela.BASE
CAB = stimela.CAB
USER = stimela.USER
UID = stimela.UID
GID = stimela.GID
LOG_HOME = stimela.LOG_HOME
LOG_FILE = stimela.LOG_FILE

class MultilineFormatter(argparse.HelpFormatter):
    def _fill_text(self, text, width, indent):
        text = self._whitespace_matcher.sub(' ', text).strip()
        paragraphs = text.split('|n ')
        multiline_text = ''
        for paragraph in paragraphs:
            formatted_paragraph = _textwrap.fill(
                paragraph, width, initial_indent=indent, subsequent_indent=indent) + '\n\n'
            multiline_text = multiline_text + formatted_paragraph
        return multiline_text

def build(argv):
    for i, arg in enumerate(argv):
        if (arg[0] == '-') and arg[1].isdigit():
            argv = ' ' + arg

    parser = ArgumentParser(description='Build executor (a.k.a cab) images')
    parser.add_argument("-b", "--base", action="store_true",
                        help="Build base images")

    parser.add_argument("-c", "--cab", metavar="CAB,CAB_DIR",
                        help="Executor image (name) name, location of executor image files")

    parser.add_argument("-uo", "--us-only",
                        help="Only build these cabs. Comma separated cab names")

    parser.add_argument("-i", "--ignore-cabs", default="",
                        help="Comma separated cabs (executor images) to ignore.")

    parser.add_argument("-nc", "--no-cache", action="store_true",
                        help="Do not use cache when building the image")

    parser.add_argument("-bl", "--build-label", default=USER,
                        help="Label for cab images. All cab images will be named <CAB_LABEL>_<cab name>. The default is $USER")

    args = parser.parse_args(argv)
    log = logger.StimelaLogger(
        '{0:s}/{1:s}_stimela_logfile.json'.format(LOG_HOME, args.build_label), jtype="docker")

    no_cache = ["--no-cache"] if args.no_cache else []

    if args.base:
        # Build base and meqtrees images first
        BASE.remove("base")
        BASE.remove("meqtrees")
        BASE.remove("casa")
        BASE.remove("astropy")

        for image in ["base", "meqtrees", "casa", "astropy"] + BASE:
            dockerfile = "{:s}/{:s}".format(stimela.BASE_PATH, image)
            image = "stimela/{0}:{1}".format(image, stimela.__version__)
            docker.build(image,
                         dockerfile, args=no_cache)

        log.log_image(image, dockerfile, replace=True)
        log.write()

        return 0

    workdir = "/home/{}/output/".format(USER)
    build_args = [
        "WORKDIR {:s}".format(workdir),
        "RUN useradd -r -u {0:d} -U {1:s}".format(UID, USER),
        "USER {0:s}".format(USER),
    ]

    if args.cab:
        cab_args = args.cab.split(",")

        if len(cab_args) == 2:
            cab, path = cab_args
        else:
            raise ValueError("Not enough arguments for build command.")

        image = "{:s}_cab/{:s}".format(args.build_label, cab)

        docker.build(image,
                     path,
                     build_args=build_args, args=no_cache)

        log.log_image(image, path, replace=True, cab=True)
        log.write()
        return

    # Cabs and their locations
    cabs = []
    dockerfiles = []

    if args.us_only:
        CABS = args.us_only.split(',')
    else:
        # Images that have been logged
        # This is crucial for making custom cabs
        logged_images = log.read().get('images', {})
        for key, val in logged_images.items():
            if val['CAB']:
                cabs.append(key)
                dockerfiles.append(val['DIR'])
        # If user wants to ignore some cabs
        IGNORE = args.ignore_cabs.split(",")
        CABS = set(CAB).difference(set(IGNORE))

    # Prioritise package images over logged images
    cabs = ["{:s}_cab/{:s}".format(args.build_label, cab)
            for cab in CABS] + cabs
    dockerfiles = ["{:s}/{:s}".format(stimela.CAB_PATH, cab)
                   for cab in CABS] + dockerfiles
    built = []
    for image, dockerfile in zip(cabs, dockerfiles):
        if image not in built:
            docker.build(image,
                         dockerfile,
                         build_args=build_args, args=no_cache)

            log.log_image(image, dockerfile, replace=True, cab=True)
            log.write()
            built.append(image)


def get_cabs(logfile):
    log = logger.StimelaLogger(logfile)
    cabs_ = log.read()['images']

    # Remove images that are not cabs
    keys = list(cabs_.keys())
    for key in keys:
        if not cabs_[key]['CAB']:
            del cabs_[key]

    return cabs_


def info(cabdir, header=False):
    """ prints out help information about a cab """

    # First check if cab exists
    pfile = "{}/parameters.json".format(cabdir)
    if not os.path.exists(pfile):
        raise RuntimeError("Cab could not be found at : {}".format(cabdir))
    # Get cab info
    cab_definition = cab.CabDefinition(parameter_file=pfile)
    cab_definition.display(header)


def cabs(argv):
    for i, arg in enumerate(argv):
        if (arg[0] == '-') and arg[1].isdigit():
            argv = ' ' + arg

    parser = ArgumentParser(description='List executor (a.k.a cab) images')
    parser.add_argument("-i", "--cab-doc",
                        help="Will display document about the specified cab. For example, \
to get help on the 'cleanmask cab' run 'stimela cabs --cab-doc cleanmask'")

    parser.add_argument("-l", "--list", action="store_true",
                        help="List cab names")

    parser.add_argument("-ls", "--list-summary", action="store_true",
                        help="List cabs with a summary of the cab")

    parser.add_argument("-bl", "--build-label", default=USER,
                        help="Label for build you want documentation for. See --build-label option in 'stimela help build'")

    args = parser.parse_args(argv)
    logfile = '{0:s}/{1:s}_stimela_logfile.json'.format(
        LOG_HOME, args.build_label)

    if args.cab_doc:
        name = '{0:s}_cab/{1:s}'.format(args.build_label, args.cab_doc)
        cabdir = "{:s}/{:s}".format(stimela.CAB_PATH, args.cab_doc)
        info(cabdir)

    elif args.list_summary:
        for val in CAB:
            cabdir = "{:s}/{:s}".format(stimela.CAB_PATH, val)
            try:
                info(cabdir, header=True)
            except IOError:
                pass
    else:
        # print them cabs
        print(', '.join(CAB))


def run(argv):
    for i, arg in enumerate(argv):
        if (arg[0] == '-') and arg[1].isdigit():
            argv[i] = ' ' + arg

    parser = ArgumentParser(description='Dockerized Radio Interferometric Scripting Framework.\n'
                                        'Sphesihle Makhathini <sphemakh@gmail.com>')

    add = parser.add_argument

    add("-in", "--input",
        help="Input folder")

    add("-out", "--output",
        help="Output folder")

    add("-ms", "--msdir",
        help="MS folder. MSs should be placed here. Also, empty MSs will be placed here")

    add("-j", "--ncores", type=int,
        help="Number of cores to when stimela parallesization (stimea.utils.pper function) ")

    add("script",
        help="Run script")

    add("-g", "--globals", metavar="KEY=VALUE[:TYPE]", action="append", default=[],
        help="Global variables to pass to script. The type is assumed to string unless specified")

    add("-bl", "--build-label", default=USER,
        help="Label for cab images. All cab images will be named <CAB_LABEL>_<cab name>. The default is $USER")

    args = parser.parse_args(argv)
    tag = None

    _globals = dict(_STIMELA_INPUT=args.input, _STIMELA_OUTPUT=args.output,
                    _STIMELA_MSDIR=args.msdir,
                    CAB_TAG=tag, _STIMELA_BUILD_LABEL=args.build_label)

    nargs = len(args.globals)

    global GLOBALS

    if nargs:
        for arg in args.globals:
            if arg.find("=") > 1:
                key, value = arg.split("=")

                try:
                    value, _type = value.split(":")
                except ValueError:
                    _type = "str"

                GLOBALS[key] = eval("{:s}('{:s}')".format(_type, value))

    if args.ncores:
        utils.CPUS = args.ncores

    with open(args.script, 'r') as stdr:
        exec(stdr.read(), _globals)


def pull(argv):
    for i, arg in enumerate(argv):
        if (arg[0] == '-') and arg[1].isdigit():
            argv[i] = ' ' + arg

    parser = ArgumentParser(description='Pull docker stimela base images')

    add = parser.add_argument

    add("-im", "--image", nargs="+", metavar="IMAGE[:TAG]",
        help="Pull base image along with its tag (or version). Can be called multiple times")

    add("-f", "--force", action="store_true",
        help="force pull if image already exists")

    add("-s", "--singularity", action="store_true",
        help="Pull base images using singularity."
        "Images will be pulled into the directory specified by the enviroment varaible, SINGULARITY_PULLFOLDER. $PWD by default")

    add("-d", "--docker", action="store_true",
        help="Pull base images using docker.")

    add("-p", "--podman", action="store_true",
        help="Pull base images using podman.")

    add("-cb", "--cab-base", nargs="+",
        help="Pull base image for specified cab")

    add("-pf", "--pull-folder",
        help="Images will be placed in this folder. Else, if the environmnental variable 'SINGULARITY_PULLFOLDER' is set, then images will be placed there. "
        "Else, images will be placed in the current directory")

    args = parser.parse_args(argv)

    if args.pull_folder:
        pull_folder = args.pull_folder
    else:
        try:
            pull_folder = os.environ["SINGULARITY_PULLFOLDER"]
        except KeyError:
            pull_folder = "."

    if args.docker:
        jtype = "docker"
    elif args.podman:
        jtype = "podman"
    elif args.singularity:
        jtype = "singularity"
    else:
        jtype = "udocker"

    log = logger.StimelaLogger(LOG_FILE, jtype=jtype)
    images = log.read()['images']

    images_ = []
    for cab in args.cab_base or []:
        if cab in CAB:
            filename = "/".join([stimela.CAB_PATH, cab, "parameters.json"])
            param = utils.readJson(filename)
            images_.append(":".join([param["base"], param["tag"]]))

    args.image = images_ or args.image
    if args.image:
        for image in args.image:
            simage = image.replace("/", "_")
            simage = simage.replace(":", "_") + ".img"
            if args.singularity:
                singularity.pull(
                    image, simage, directory=pull_folder, force=args.force)
            elif args.docker:
                docker.pull(image)
                log.log_image(image, 'pulled')
            elif args.podman:
                podman.pull(image)
                log.log_image(image, 'pulled')
            else:
                udocker.pull(image)
                log.log_image(image, 'pulled')
    else:

        base = []
        for cab in CAB:
            image = "{:s}/{:s}".format(stimela.CAB_PATH, cab)
            base.append(utils.get_Dockerfile_base_image(image).split()[-1])

        base = set(base)

        for image in base:
            if args.singularity:
                simage = image.replace("/", "_")
                simage = simage.replace(":", "_") + ".img"
                singularity.pull(
                    image, simage, directory=pull_folder, force=args.force)
            elif args.docker:
                docker.pull(image, force=args.force)
                log.log_image(image, 'pulled')
            elif args.podman:
                podman.pull(image, force=args.force)
                log.log_image(image, 'pulled')
            else:
                udocker.pull(image, force=args.force)
                log.log_image(image, 'pulled')

    log.write()


def images(argv):
    for i, arg in enumerate(argv):
        if (arg[0] == '-') and arg[1].isdigit():
            argv[i] = ' ' + arg

    parser = ArgumentParser(description='List all stimela related images.')

    add = parser.add_argument

    add("-c", "--clear", action="store_true",
        help="Clear the logfile that keeps track of stimela images. This does not do anythig to the images.")

    args = parser.parse_args(argv)

    log = logger.StimelaLogger(LOG_FILE)
    log.display('images')

    if args.clear:
        log.clear('images')
        log.write()


def containers(argv):
    for i, arg in enumerate(argv):
        if (arg[0] == '-') and arg[1].isdigit():
            argv[i] = ' ' + arg

    parser = ArgumentParser(description='List all active stimela containers.')

    add = parser.add_argument

    add("-c", "--clear", action="store_true",
        help="Clear the log file that keeps track of stimela containers. This doesn't do anything to the containers.")

    args = parser.parse_args(argv)

    log = logger.StimelaLogger(LOG_FILE)
    log.display('containers')
    if args.clear:
        log.clear('containers')
        log.write()


def ps(argv):
    for i, arg in enumerate(argv):
        if (arg[0] == '-') and arg[1].isdigit():
            argv[i] = ' ' + arg

    parser = ArgumentParser(description='List all running stimela processes')

    add = parser.add_argument

    add("-c", "--clear", action="store_true",
        help="Clear logfile that keeps track of stimela processes. This doesn't do anything ot the processes themselves.")

    args = parser.parse_args(argv)

    log = logger.StimelaLogger(LOG_FILE)
    log.display('processes')
    if args.clear:
        log.clear('processes')
        log.write()


def kill(argv):
    for i, arg in enumerate(argv):
        if (arg[0] == '-') and arg[1].isdigit():
            argv[i] = ' ' + arg

    parser = ArgumentParser(description='Gracefully kill stimela process(s).')

    add = parser.add_argument

    add("pid", nargs="*",
        help="Process ID")

    args = parser.parse_args(argv)

    log = logger.StimelaLogger(LOG_FILE)

    for pid in args.pid:

        found = pid in log.info['processes'].keys()

        if not found:
            print("Could not find process {0}".format(pid))
            continue

        try:
            os.kill(int(pid), signal.SIGINT)
        except OSError:
            raise OSError(
                'Process with PID {} could not be killed'.format(pid))

        log.remove('processes', pid)
    log.write()


def clean(argv):
    for i, arg in enumerate(argv):
        if (arg[0] == '-') and arg[1].isdigit():
            argv[i] = ' ' + arg

    parser = ArgumentParser(
        description='Convience tools for cleaning up after stimela')
    add = parser.add_argument

    add("-ai", "--all-images", action="store_true",
        help="Remove all images pulled/built by stimela. This include CAB images")

    add("-ab", "--all-base", action="store_true",
        help="Remove all base images")

    add("-ac", "--all-cabs", action="store_true",
        help="Remove all CAB images")

    add("-aC", "--all-containers", action="store_true",
        help="Stop and/or Remove all stimela containers")

    add("-bl", "--build-label", default=USER,
        help="Label for cab images. All cab images will be named <CAB_LABEL>_<cab name>. The default is $USER")

    args = parser.parse_args(argv)

    log = logger.StimelaLogger(LOG_FILE)
    log_cabs = logger.StimelaLogger('{0:s}/{1:s}_stimela_logfile.json'.format(LOG_HOME,
                                                                              args.build_label))

    if args.all_images:
        images = log.info['images'].keys()
        images = log_cabs.info['images'].keys()
        for image in images:
            utils.xrun('docker', ['rmi', image])
            log.remove('images', image)
            log.write()

        images = log_cabs.info['images'].keys()
        for image in images:
            if log_cabs.info['images'][image]['CAB']:
                utils.xrun('docker', ['rmi', image])
                log_cabs.remove('images', image)
                log_cabs.write()

    if args.all_base:
        images = log.info['images'].keys()
        for image in images:
            if log.info['images'][image]['CAB'] is False:
                utils.xrun('docker', ['rmi', image])
                log.remove('images', image)
                log.write()

    if args.all_cabs:
        images = log_cabs.info['images'].keys()
        for image in images:
            if log_cabs.info['images'][image]['CAB']:
                utils.xrun('docker', ['rmi', image])
                log_cabs.remove('images', image)
                log_cabs.write()

    if args.all_containers:
        containers = log.info['containers'].keys()
        for container in containers:
            cont = docker.Container(
                log.info['containers'][container]['IMAGE'], container)
            try:
                status = cont.info()['State']['Status'].lower()
            except:
                print('Could not inspect container {}. It probably doesn\'t exist, will remove it from log'.format(
                    container))
                status = "no there"

            if status == 'running':
                # Kill the container instead of stopping it, so that effect can be felt py parent process
                utils.xrun('docker', ['kill', container])
                cont.remove()
            elif status in ['exited', 'dead']:
                cont.remove()

            log.remove('containers', container)
            log.write()


def main(argv):
    for i, arg in enumerate(argv):
        if (arg[0] == '-') and arg[1].isdigit():
            argv[i] = ' ' + arg

    parser = ArgumentParser(description='Stimela: Dockerized Radio Interferometric Scripting Framework. '
                            '|n version {:s} |n install path {:s} |n '
                            'Sphesihle Makhathini <sphemakh@gmail.com>'.format(stimela.__version__,
                                                                               os.path.dirname(__file__)),
                            formatter_class=MultilineFormatter,
                            add_help=False)

    add = parser.add_argument

    add("-h", "--help",  action="store_true",
        help="Print help message and exit")

    add("-v", "--version", action='version',
        version='{:s} version {:s}'.format(parser.prog, stimela.__version__))

    add("command", nargs="*", metavar="command [options]",
        help="Stimela command to execute. For example, 'stimela help run'")

    options = []
    commands = dict(pull=pull, build=build, run=run,
                    images=images, cabs=cabs, ps=ps,
                    containers=containers, kill=kill,
                    clean=clean)

    command = "failure"

    for cmd in commands:
        if cmd in argv:
            command = cmd

            index = argv.index(cmd)
            options = argv[index+1:]
            argv = argv[:index+1]

    args = parser.parse_args(argv)

    # Command is help and no other commands following
    try:
        main_help = (args.command[0] == "help" and len(args.command) == 1)
    except IndexError:
        main_help = True

    if args.help or main_help:
        parser.print_help()

        print("""
Run a command. These can be:

help    : Prints out a help message about other commands
build   : Build a set of stimela images
pull    : pull a stimela base images
run     : Run a stimela script
images  : List stimela images
cabs    : Manage cab images
ps      : List running stimela scripts
kill    : Gracefully kill runing stimela process
clean   : Clean up tools for stimela

""")

        raise SystemExit

    # Separate commands into command and arguments
    cmd, argv = args.command[0], args.command[1:]

    # If we've got past the if statement above, and help
    # is the command then assume that help on a command
    # is requested
    if cmd == "help":
        # Request help on the sub-command
        cmd, argv = argv[0], ["-h"]
    else:
        argv = options

    # Get the function to execute for the command
    try:
        _cmd = commands[cmd]
    except KeyError:
        raise KeyError("Command '{:s}' not recognized "
                       "Run : 'stimela help' for help".format(cmd))

    # Invoke the command
    _cmd(argv)



import os
import sys

sys.path.append('/scratch/stimela')

utils = __import__('utils')


CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
MSDIR = os.environ["MSDIR"]
OUTPUT = os.environ["OUTPUT"]

cab = utils.readJson(CONFIG)
args = []
targets = None
for param in cab['parameters']:
    name = param['name']
    value = param['value']

    if value is None:
        continue
    elif value is False:
        continue
    elif value is True:
        value = ''
    elif name == 'target-images':
        targets = value
        continue

    args += ['{0}{1} {2}'.format(cab['prefix'], name, value)]

indir = os.path.dirname(targets[0])
target_names = map(os.path.basename, targets)
target_images = "--target-images " + " --target-images ".join(target_names)

args += ["--input {0:s} {1:s} --output {2:s}".format(indir, target_images,
                                                     OUTPUT)]

if target_images:
    utils.xrun(cab['binary'], args)
else:
    raise RuntimeError(
        'Filenames of the images to be mosaicked have not been specified.')



import subprocess
import os
import sys
from io import StringIO
from stimela import utils
import json
import stimela
import time
import datetime
import tempfile


class DockerError(Exception):
    pass


def pull(image, tag=None, force=False):
    """ pull a docker image """
    if tag:
        image = ":".join([image, tag])

    utils.xrun("udocker", ["pull", image])


def seconds_hms(seconds):
    return str(datetime.timedelta(seconds=seconds))


class Container(object):
    def __init__(self, image, name,
                 volumes=None, environs=None,
                 label="", logger=None,
                 shared_memory="1gb",
                 time_out=-1,
                 log_container=None,
                 COMMAND="",
                 use_graphics=False):
        """
        Python wrapper to docker engine tools for managing containers.
        """

        self.image = image
        self.name = name
        self.label = label
        self.volumes = volumes or []
        self.environs = environs or []
        self.logger = logger
        self.status = None
        self.WORKDIR = None
        self.COMMAND = COMMAND
        self.PID = os.getpid()
        self.uptime = "00:00:00"
        self.time_out = time_out
        self.use_graphics = use_graphics
        self.cont_logger = utils.logger.StimelaLogger(
            log_container or stimela.LOG_FILE, jtype="udocker")

    def add_volume(self, host, container, noverify=False):

        if os.path.exists(host) or noverify:
            if self.logger:
                self.logger.debug("Mounting volume [{0}] in container [{1}] at [{2}]".format(
                    host, self.name, container))
            host = os.path.abspath(host)
        else:
            raise IOError(
                "Directory {0} cannot be mounted on container: File doesn't exist".format(host))

        self.volumes.append(":".join([host, container]))

    def add_environ(self, key, value):
        if self.logger:
            self.logger.debug("Adding environ varaible [{0}={1}] in container {2}".format(
                key, value, self.name))
        self.environs.append('"{0:s}={1:s}"'.format(key, value))

    def run(self, *args):

        if self.volumes:
            volumes = " --volume=" + " --volume=".join(self.volumes)
        else:
            volumes = ""
        if self.environs:
            environs = environs = " --env="+" --env=".join(self.environs)
        else:
            environs = ""

        self._print("Running container [{}]".format(self.name))
        tstart = time.time()
        utils.xrun("udocker run", ["=".join(args)] + [volumes, environs,
                                                      "--workdir=%s" % (
                                                          self.WORKDIR) if self.WORKDIR else "",
                                                      "--rm",
                                                      "--dri" if self.use_graphics else "",
                                                      self.name, self.COMMAND or ""], timeout=self.time_out)

        self.status = "running"
        uptime = seconds_hms(time.time() - tstart)
        self.uptime = uptime
        self._print("Runtime was {0}.".format(uptime))

    def info(self):

        output = subprocess.check_output(
            "udocker inspect {}".format(self.name), shell=True).decode()
        output_file = StringIO(output[3:-3])
        jdict = json.load(output_file)
        output_file.close()

        return jdict

    def get_log(self):
        stdout = open(self.logfile, 'w')
        exit_status = subprocess.call("docker logs {0}".format(self.name),
                                      stdout=stdout, stderr=stdout, shell=True)
        if exit_status != 0:
            self.logger.warn(
                'Could not log container: {}. Something went wrong durring execution'.format(self.name))
            output = 'Task was not started.'
            stdout.write(output)
        else:
            output = stdout.read()

        stdout.close()
        return output

    def create(self):
        running = True
        self.status = "created"

        # self.cont_logger.log_container(self.name)
        # self.cont_logger.write()
        self._print("Creating container [{0:s}]. Timeout set to {1:d}. The container ID is printed below.".format(
            self.name, self.time_out))
        utils.xrun(
            "udocker", ["create", "--name={0:s}".format(self.name), self.image])

    def stop(self):
        dinfo = self.info()
        status = dinfo["State"]["Status"]
        killed = False
        if status in ["running", "paused"]:
            try:
                utils.xrun("docker stop", [self.name])
            except KeyboardInterrupt("Received terminate signal. Will stop and remove container first"):
                killed = True
        self.status = 'exited'

        self._print("Container {} has been stopped.".format(self.name))
        if killed:
            self.remove()
            raise KeyboardInterrupt

    def remove(self):
        dinfo = self.info()
        status = dinfo["State"]["Status"]
        killed = False
        if status == "exited":
            try:
                utils.xrun("udocker rm", [self.name])
            except KeyboardInterrupt:
                killed = True
            if killed:
                raise KeyboardInterrupt

        else:
            raise DockerError(
                "Container [{}] has not been stopped, cannot remove".format(self.name))

#        self.cont_logger.remove('containers', self.name)
#        self.cont_logger.write()

    def _print(self, message):
        if self.logger:
            self.logger.info(message)
        else:
            print(message)



#!/usr/bin/env python

import os
import sys
from setuptools import setup


requirements = ["pyyaml",
                "nose>=1.3.7",
                "future",
                ],

if sys.version_info <= (3, 0):
    requirements += [
        "udocker",
    ],

PACKAGE_NAME = "stimela"
__version__ = "1.2.1"

setup(name=PACKAGE_NAME,
      version=__version__,
      description="Dockerized radio interferometry scripting framework",
      author="Sphesihle Makhathini",
      author_email="sphemakh@gmail.com",
      url="https://github.com/sphemakh/Stimela",
      packages=["stimela", "stimela_misc", "stimela/cargo",
                "stimela/utils", "stimela/cargo/cab",
                "stimela/cargo/base"],
      package_data={"stimela/cargo": [
          "cab/*/Dockerfile",
          "base/*/Dockerfile",
          "cab/*/src/*.py",
          "cab/*/src/*.sh",
          "cab/*/src/*.json",
          "base/*/xvfb.init.d",
          "cab/*/xvfb.init.d",
          "cab/*/parameters.json",
          "cab/*/src/tdlconf.profiles",
          "cab/singularity_run",
          "cab/docker_run",
      ]},
      install_requires=requirements,
      scripts=["bin/" + i for i in os.listdir("bin")],
      classifiers=[],
      )



import os
import sys

sys.path.append('/scratch/stimela')

utils = __import__('utils')

CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
MSDIR = os.environ["MSDIR"]

cab = utils.readJson(CONFIG)
args = []

for param in cab['parameters']:
    name = param['name']
    value = param['value']
    if value is None:
        continue
    elif value is False:
        continue

    if type(value) is list:
        args += ['{0}{1} {2}'.format(cab['prefix'], name,
                                     " ".join([val.replace(':output', '') for val in value]))]
    else:
        args += ['{0}{1} {2}'.format(cab['prefix'], name, value)]

utils.xrun(cab['binary'], args)



import os
import sys

sys.path.append('/scratch/stimela')

utils = __import__('utils')

CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
OUTPUT = os.environ["OUTPUT"]
MSDIR = os.environ["MSDIR"]

cab = utils.readJson(CONFIG)

args = []
for param in cab['parameters']:
    name = param['name']
    value = param['value']

    if value in [False, None]:
        continue

    if name in 'channels timeslots corrs stations ddid field str'.split():
        if name in 'channels timeslots'.split():
            value = ':'.join(value)
        else:
            value = ','.join(value)
    if value is True:
        value = ""

    if name == 'msname':
        msname = value
        if isinstance(msname, str):
            msname = [msname]
        continue

    if name == 'flagged-any':
        args += ['{0}flagged-any {1}'.format(cab['prefix'], a) for a in value]
        continue

    args.append('{0}{1} {2}'.format(cab['prefix'], name, value))

utils.xrun(cab['binary'], args+msname)



import numpy
import os
import sys
from pyrap.tables import table
import subprocess

sys.path.append("/scratch/stimela")

utils = __import__('utils')

CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
OUTPUT = os.environ["OUTPUT"]
MSDIR = os.environ["MSDIR"]
CODE = "/scratch/code"

cab = utils.readJson(CONFIG)
binary = cab['binary']
parameters = cab['parameters']

jdict = {}
for param in parameters:
    name = param['name']
    value = param['value']
    if value is None:
        continue
    if value is False:
        value = 0
    elif value is True:
        value = 1

    jdict[name] = value

msname = jdict['msname']
THREADS = jdict.pop('threads', 2)
msbase = os.path.basename(msname)[:-3]
prefix = jdict.pop('prefix', None) or '{0}/{1}'.format(OUTPUT, msbase)
params = {}

# options for writing flags
writeflags = jdict.pop("write-flags-to-ms", None)
if writeflags:
    params["ms_sel.ms_write_flags"] = 1
    params["ms_sel.ms_fill_legacy_flags"] = 1 if jdict.pop(
        "fill-legacy-flags", False) else 0

write_flagset = jdict.pop("write-flagset", None)
if write_flagset:
    params["ms_wfl.write_bitflag"] = write_flagset
    params["ms_sel.ms_write_flag_policy"] = "'add to set'" if jdict.pop(
        "write-flag-policy", False) else "'replace set'"

# Read flags options
readflagsets = jdict.pop("read-flagsets", False)
if readflagsets:
    params["ms_rfl.read_flagsets"] = readflagsets

params['ms_sel.ms_read_flags'] = 1 if jdict.pop(
    "read-flags-from-ms", False) else 0
params["ms_rfl.read_legacy_flags"] = 1 if jdict.pop(
    "read-legacy-flags", False) else 0

params["ms_sel.msname"] = msname
field_id = jdict.pop("field-id", 0)
spw_id = jdict.pop("spw-id", 0)
params["ms_sel.tile_size"] = jdict.pop("tile-size", 16)
params["ms_sel.ddid_index"] = spw_id
params["ms_sel.field_index"] = field_id

TDL = jdict.pop("tdlconf", None) or CODE + "/tdlconf.profiles"
SECTION = jdict.pop("section", None) or "stefcal"
skymodel = jdict.pop("skymodel", None)
beam_files_pattern = jdict.pop("beam-files-pattern", False)
jones_type = jdict.pop("jones-implementation", "Gain2x2")

column = jdict.pop("column", "DATA")
outcol = jdict.pop("output-column", "CORRECTED_DATA")

params["ms_sel.input_column"] = column
params["ms_sel.output_column"] = outcol
params["tiggerlsm.filename"] = skymodel
params["tiggerlsm.lsm_subset"] = jdict.get("subset", "all")
params["do_output"] = jdict.pop("output-data", "CORR_RES")
saveconf = jdict.pop('save-config', None)
params['ms_sel.ms_corr_sel'] = "'{}'".format(jdict.pop('correlations', '2x2'))

label = jdict.pop("label", None)

model_column = jdict.pop("model-column", 'MODEL_DATA')

gjones = jdict.pop("Gjones", False)
if gjones:

    time_smooth, freq_smooth = jdict.get("Gjones-smoothing-intervals", (1, 1))
    time_int, freq_int = jdict.get("Gjones-solution-intervals", (1, 1))
    mode = 'apply' if jdict.get('Gjones-apply-only', False) else 'solve-save'

    gjones_gains = jdict.pop('Gjones-gain-table', None) or "{0}/{1}{2}.gain.cp".format(
        OUTPUT, msbase, "-%s" % label if label else "")
    params.update({
        "stefcal_gain.enabled": 1,
        "stefcal_gain.mode": mode,
        "stefcal_gain.reset": 0 if mode == "apply" else 1,
        "stefcal_gain.implementation": jones_type,
        "stefcal_gain.timeint": time_int,
        "stefcal_gain.freqint": freq_int,
        "stefcal_gain.flag_ampl":   jdict.get("Gjones-ampl-clipping", 0),
        "stefcal_gain.flag_chisq":   jdict.get("Gjones-chisq-clipping", 0),
        "stefcal_gain.flag_chisq_threshold":   jdict.get("Gjones-thresh-sigma", 10),
        "stefcal_gain.flag_ampl_low":   jdict.get("Gjones-ampl-clipping-low", 0.3),
        "stefcal_gain.flag_ampl_high":   jdict.get("Gjones-ampl-clipping-high", 2),
        "stefcal_gain.implementation":   jdict.get("Gjones-matrix-type", "Gain2x2"),
        "stefcal_gain.table": gjones_gains,
    })

bjones = jdict.pop("Bjones", False)
if bjones:

    time_smooth, freq_smooth = jdict.get("Bjones-smoothing-intervals", (1, 0))
    time_int, freq_int = jdict.get("Bjones-solution-intervals", (1, 0))
    mode = 'apply' if jdict.get('Bjones-apply-only', False) else 'solve-save'

    bjones_gains = jdict.pop('Bjones-gain-table', None) or "{0}/{1}{2}.gain1.cp".format(
        OUTPUT, msbase, "-%s" % label if label else "")
    params.update({
        "stefcal_gain1.enabled": 1,
        "stefcal_gain1.label": 'B',
        "stefcal_gain1.mode": mode,
        "stefcal_gain1.reset": 0 if mode == "apply" else 1,
        "stefcal_gain1.implementation": jones_type,
        "stefcal_gain1.timeint": time_int,
        "stefcal_gain1.freqint": freq_int,
        "stefcal_gain1.flag_ampl":   jdict.get("Bjones-ampl-clipping", 0),
        "stefcal_gain1.flag_chisq":   jdict.get("Bjones-chisq-clipping", 0),
        "stefcal_gain1.flag_chisq_threshold":   jdict.get("Bjones-thresh-sigma", 10),
        "stefcal_gain1.flag_ampl_low":   jdict.get("Bjones-ampl-clipping-low", 0.3),
        "stefcal_gain1.flag_ampl_high":   jdict.get("Bjones-ampl-clipping-high", 2),
        "stefcal_gain1.implementation":   jdict.get("Bjones-matrix-type", "Gain2x2"),
        "stefcal_gain1.table": bjones_gains,
    })

beam = jdict.pop("Ejones", False)
if beam and beam_files_pattern:
    params.update({
        "me.e_enable": 1,
        "me.p_enable": 1,
        "me.e_module": "Siamese_OMS_pybeams_fits",
        "me.e_all_stations": 1,
        "pybeams_fits.sky_rotation":  1 if jdict.pop('parallactic-angle-rotation', False) else 0,
        "pybeams_fits.l_axis": jdict.pop("beam-l-axis", "L"),
        "pybeams_fits.m_axis": jdict.pop("beam-m-axis", "M"),
        "pybeams_fits.filename_pattern": "'{}'".format(beam_files_pattern),
    })


ddjones = jdict.pop("DDjones", False)
if ddjones:
    time_int, freq_int = jdict.pop("DDjones-solution-intervals", (1, 1))
    time_smooth, freq_smooth = jdict.pop("DDjones-smoothing-intervals", (1, 1))

    mode = 'apply' if jdict.get('DDjones-apply-only', False) else 'solve-save'

    ddjones_gains = jdict.pop('DDjones-gain-table', None) or "{0}/{1}{2}.diffgain.cp".format(
        OUTPUT, msbase, "-%s" % label if label else "")
    params.update({
        "stefcal_diffgain.enabled": 1,
        "stefcal_diffgain.reset": 0 if mode == "apply" else 1,
        "stefcal_diffgain.flag_ampl": 0,
        "stefcal_diffgain.flag_chisq": 1,
        "stefcal_diffgain.flag_chisq_threshold": 5,
        "stefcal_diffgain.freqint": freq_int,
        "stefcal_diffgain.freqsmooth": freq_smooth,
        "stefcal_diffgain.implementation": jones_type,
        "stefcal_diffgain.label": jdict.pop("DDjones-tag", "dE"),
        "stefcal_diffgain.max_diverge": 1,
        "stefcal_diffgain.mode": mode,
        "stefcal_diffgain.niter": 50,
        "stefcal_diffgain.omega": 0.5,
        "stefcal_diffgain.quota": 0.95,
        "stefcal_diffgain.table": ddjones_gains,
        "stefcal_diffgain.timeint": time_int,
        "stefcal_diffgain.timesmooth": time_smooth,
        "stefcal_diffgain.implementation":   jdict.get("DDjones-matrix-type", "Gain2x2"),
    })

ifrjones = jdict.pop("DDjones", False)
if ifrjones:
    ifrjones_gains = jdict.pop('IFRjones-gain-table', None) or "{0}/{1}{2}.ifrgain.cp".format(
        OUTPUT, msbase, "-%s" % label if label else "")
    mode = 'apply' if jdict.get('IFRjones-apply-only', False) else 'solve-save'

    params.update({
        "stefcal_ifr_gain_mode": mode,
        "stefcal_ifr_gains": 1,
        "stefcal_ifr_gain_reset": 0 if mode == "apply" else 1,
        "stefcal_reset_ifr_gains": 0 if mode == "apply" else 1,
        "stefcal_ifr_gain.table": ifrjones_gains,
    })

makeplots = jdict.pop("make-plots", False)

gjones_plotprefix = prefix+"-gjones_plots"
bjones_plotprefix = prefix+"-bjones_plots"
ddjones_plotprefix = prefix+"-ddjones_plots"
ifrjones_plotprefix = prefix+"-ifrjones_plots"


def run_meqtrees(msname):

    prefix = ["--mt %d -c %s [%s]" % (THREADS, TDL, SECTION)]
    suffix = ["%s/Calico/calico-stefcal.py =stefcal" %
              os.environ["MEQTREES_CATTERY_PATH"]]
    options = {}
    options.update(params)
    if jdict.pop("add-vis-model", 0):
        options["read_ms_model"] = 1
        options["ms_sel.model_column"] = model_column

    taql = jdict.get('data-selection', None)
    if taql:
        options["ms_sel.ms_taql_str"] = taql

    args = []
    for key, value in options.iteritems():
        if isinstance(value, str) and value.find(' ') > 0:
            value = '"{:s}"'.format(value)
        args.append('{0}={1}'.format(key, value))

    args = prefix + args + suffix

    utils.xrun(cab['binary'], args +
               ['-s {}'.format(saveconf) if saveconf else ''])

    print("MeqTrees Done!")
    # now plot the gains
    if makeplots:
        print("Preparing to make gain plots")
        import Owlcat.Gainplots as plotgains
        feed_tab = table(msname+"/FEED")
        print("Extracting feed type from MS")
        feed_type = set(feed_tab.getcol("POLARIZATION_TYPE")['array'])
        feed_type = "".join(feed_type)
        print("Feed type is [%s]" % feed_type)

        if feed_type.upper() in ["XY", "YX"]:
            feed_type = "XY"
        else:
            feed_type = "RL"

        if gjones:
            print("Making Gain plots (G)...")
            plotgains.make_gain_plots(
                gjones_gains, prefix=gjones_plotprefix, feed_type=feed_type)

        if bjones:
            print("Making Gain plots (B)...")
            plotgains.make_gain_plots(
                bjones_gains, gain_label='B', prefix=bjones_plotprefix, feed_type=feed_type)

        if ddjones:
            print("Making differential gain plots...")
            plotgains.make_diffgain_plots(
                ddjones_gains, prefix=ddjones_plotprefix, feed_type=feed_type)

        if ifrjones:
            print("Making IFR gain plots...")
            plotgains.make_ifrgain_plots(
                ifrjones_gains, prefix=ifrjones_plotprefix, feed_type=feed_type)

    sys.exit(0)


run_meqtrees(msname)
